# ----------------------------------------------------------------------------
# cocos2d
# Copyright (c) 2008 Daniel Moisset, Ricardo Quesada, Rayentray Tappa, Lucio Torre
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
"""
Scene class and subclasses
"""
__docformat__ = 'restructuredtext'

__all__ = ['Scene']

from pyglet.gl import *

import cocos
from cocos.director import director
import cocos.cocosnode as cocosnode

class EventHandlerMixin(object):
    def add(self, child, *args, **kwargs):
        super(EventHandlerMixin, self).add(child, *args, **kwargs)
        
        scene = self.get(Scene)
        if not scene: return
        
        if (    scene._handlers_enabled and 
                scene.is_running and 
                isinstance(child, cocos.layer.Layer) 
                ):
            child.push_all_handlers()
            
            
    def remove(self, child):
        super(EventHandlerMixin, self).remove(child)
        
        scene = self.get(Scene)
        if not scene: return
        
        if (    scene._handlers_enabled and 
                scene.is_running and 
                isinstance(child, cocos.layer.Layer) 
                ):
            child.remove_all_handlers()
            
    
            
 
class Scene(cocosnode.CocosNode, EventHandlerMixin):
    """
    """
   
    def __init__(self, *children):
        """
        Creates a Scene with layers and / or scenes.
        
        Responsibilities:
            Control the dispatching of events to its layers
            
        :Parameters:
            `children` : list of `Layer` or `Scene`
                Layers or Scenes that will be part of the scene.
                They are automatically assigned a z-level from 0 to
                num_children.
        """

        super(Scene,self).__init__()
        self._handlers_enabled = False
        for i, c in enumerate(children):
            self.add( c, z=i )
        
        x,y = director.get_window_size()
        
        self.transform_anchor_x = x/2
        self.transform_anchor_y = y/2
            
    def on_enter(self):
        for c in self.get_children():
            c.parent = self
        super(Scene, self).on_enter()
        
            
    def push_all_handlers(self):
        for child in self.get_children():
            if isinstance(child, cocos.layer.Layer):
                child.push_all_handlers()
            
    def remove_all_handlers(self):
        for child in self.get_children():
            if isinstance(child, cocos.layer.Layer):
                child.remove_all_handlers()
    
    def enable_handlers(self, value=True):
        """
        This function makes the scene elegible for receiving events
        """
        if value and not self._handlers_enabled and self.is_running:
            self.push_all_handlers()
        elif not value and self._handlers_enabled and self.is_running:
            self.remove_all_handlers()
        self._handlers_enabled = value
        
        
    def end(self, value=None):
        """Ends the current scene setting director.return_value with `value`
        
        :Parameters:
            `value` : anything
                The return value. It can be anything. A type or an instance.
        """
        director.return_value = value
        director.pop()




# ----------------------------------------------------------------------------
# cocos2d
# Copyright (c) 2008 Daniel Moisset, Ricardo Quesada, Rayentray Tappa, Lucio Torre
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
'''Text cocos nodes
'''
__docformat__ = 'restructuredtext'

from director import director
import cocosnode
from batch import *

import pyglet
from pyglet.graphics import OrderedGroup
from pyglet import image
from pyglet.gl import *

from batch import *

class TextElement(cocosnode.CocosNode):
    def __init__(self, text='', position=(0,0), **kwargs):
        super(TextElement, self).__init__()
        self.position = position
        self.args = []
        self.kwargs = kwargs
        kwargs['text']=text
        self.group = None
        self.batch = None

        self.batch = pyglet.graphics.Batch()
        self.create_element()

    def create_element(self):
        self.element = self.klass(group=self.group, batch=self.batch, **self.kwargs)

    def draw(self):
        glPushMatrix()
        self.transform()
        self.element.draw()
        glPopMatrix()

    def _get_opacity(self):
        return self.element.color[3]
    def _set_opacity(self, value):
        self.element.color = tuple(self.element.color[:3]) + (int(value),)
    opacity = property(_get_opacity, _set_opacity)

class Label(TextElement):
    '''CocosNode Label element. It is a wrapper of pyglet.text.Label with the benefits
    of being of a CocosNode
    '''
    klass = pyglet.text.Label



class HTMLLabel(TextElement):
    '''CocosNode HTMLLabel element. It is a wrapper of pyglet.text.HTMLLabel with the benefits
    of being of a CocosNode
    '''
    klass = pyglet.text.HTMLLabel

class PygletRichLabel(pyglet.text.DocumentLabel):
    '''Rich text label.
    '''
    def __init__(self, text='',
                 font_name=None, font_size=None, bold=False, italic=False,
                 color=None,
                 x=0, y=0, width=None, height=None,
                 anchor_x='left', anchor_y='baseline',
                 halign='left',
                 multiline=False, dpi=None, batch=None, group=None):
        '''Create a rich text label.

        :Parameters:
            `text` : str
                Pyglet attributed (rich) text to display.
            `font_name` : str or list
                Font family name(s).  If more than one name is given, the
                first matching name is used.
            `font_size` : float
                Font size, in points.
            `bold` : bool
                Bold font style.
            `italic` : bool
                Italic font style.
            `color` : (int, int, int, int) or None
                Font colour, as RGBA components in range [0, 255].
                None to use font colors defined by text attributes.
            `x` : int
                X coordinate of the label.
            `y` : int
                Y coordinate of the label.
            `width` : int
                Width of the label in pixels, or None
            `height` : int
                Height of the label in pixels, or None
            `anchor_x` : str
                Anchor point of the X coordinate: one of ``"left"``,
                ``"center"`` or ``"right"``.
            `anchor_y` : str
                Anchor point of the Y coordinate: one of ``"bottom"``,
                ``"baseline"``, ``"center"`` or ``"top"``.
            `halign` : str
                Horizontal alignment of text on a line, only applies if
                a width is supplied. One of ``"left"``, ``"center"``
                or ``"right"``.
            `multiline` : bool
                If True, the label will be word-wrapped and accept newline
                characters.  You must also set the width of the label.
            `dpi` : float
                Resolution of the fonts in this layout.  Defaults to 96.
            `batch` : `Batch`
                Optional graphics batch to add the label to.
            `group` : `Group`
                Optional graphics group to use.

        '''

        text = '{color (255, 255, 255, 255)}' + text
        document = pyglet.text.decode_attributed(text)
        super(PygletRichLabel, self).__init__(document, x, y, width, height,
                                    anchor_x, anchor_y,
                                    multiline, dpi, batch, group)
        style = dict(halign=halign)

        if font_name:
            style['font_name'] = font_name
        if font_size:
            style['font_size'] = font_size
        if bold:
            style['bold'] = bold
        if italic:
            style['italic'] = italic
        if color:
            style['color'] = color

        self.document.set_style(0, len(self.document.text), style)

class RichLabel(TextElement):
    '''CocosNode RichLabel element. It is a wrapper of a custom Pyglet Rich Label
    using rich text attributes with the benefits of being of a CocosNode
    '''
    klass = PygletRichLabel


#
# cocos2d
# http://cocos2d.org
#
# This code is so you can run the samples without installing the package
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))


import cocos
from cocos.director import director

from pyglet import gl

# Defining a new layer type...
class Square(cocos.layer.Layer):
    """Square (color, c, y, size=50) : A layer drawing a square at (x,y) of
    given color and size"""
    def __init__(self, color, x, y, size=50):
        super( Square, self ).__init__()

        self.x = x
        self.y = y
        self.size = size
        self.layer_color = color
                
    def draw(self):
        super(Square,self).draw()

        gl.glColor4f(*self.layer_color)
        x, y = self.x, self.y
        w = x+self.size; h=y+self.size
        gl.glBegin(gl.GL_QUADS)
        gl.glVertex2f( x, y )
        gl.glVertex2f( x, h )
        gl.glVertex2f( w, h )
        gl.glVertex2f( w, y )
        gl.glEnd()
        gl.glColor4f(1,1,1,1) 
        
if __name__ == "__main__":
    director.init()
    # Create a large number of layers
    layers = [ Square((0.03*i,0.03*i,0.03*i,1) , i*20, i*20) for i in range(5,20) ]
    # Create a scene with all those layers
    sc = cocos.scene.Scene(*layers)
    # You can also add layers to a scene later:
    sc.add( Square((1,0,0,0.5), 150,150, 210 ), name="big_one" )
    director.run( sc )



# This code is so you can run the samples without installing the package
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../..'))

# pyglet realted
from pyglet.window import key

# cocos2d related
from cocos.layer import Layer, ColorLayer
from cocos.director import director
from cocos.text import Label
from cocos.actions import *

# tetrico related
import soundex
import hiscore
import status


class GameOver( ColorLayer ):
    is_event_handler = True #: enable pyglet's events

    def __init__( self, win = False):
        super(GameOver,self).__init__( 32,32,32,64)

        w,h = director.get_window_size()

        if win:
            soundex.play('oh_yeah.mp3')
            msg = 'YOU WIN'
        else:
            soundex.play('no.mp3')
            msg = 'GAME OVER'

        label = Label(msg,
                    font_name='Edit Undo Line BRK',
                    font_size=54,
                    anchor_y='center',
                    anchor_x='center' )
        label.position =  ( w/2.0, h/2.0 )

        self.add( label )

        angle = 5
        duration = 0.05
        accel = 2
        rot = Accelerate(Rotate( angle, duration//2 ), accel)
        rot2 = Accelerate(Rotate( -angle*2, duration), accel)
        effect = rot + (rot2 + Reverse(rot2)) * 4 + Reverse(rot)
        
        label.do( Repeat( Delay(5) + effect ) )

        if hiscore.hiscore.is_in( status.status.score ):
            self.hi_score = True

            label = Label('Enter your name:',
                        font_name='Edit Undo Line BRK',
                        font_size=36,
                        anchor_y='center',
                        anchor_x='center',
                        color=(32,32,32,255),
                        )
            label.position =  ( w/2.0, h/2.0 )
            label.position = (w//2, 300)
            self.add( label )

            self.name= Label('',
                        font_name='Edit Undo Line BRK',
                        font_size=36,
                        anchor_y='center',
                        anchor_x='center',
                        color=(32,32,32,255),
                        )
            self.name.position=(w//2,250)
            self.add(self.name)
        else:
            self.hi_score = False

    def on_key_press( self, k, m ):
        if not self.hi_score and (k == key.ENTER or k == key.ESCAPE):
            director.pop()
            return True

        if self.hi_score:
            if k == key.BACKSPACE:
                self.name.element.text = self.name.element.text[0:-1]
                return True
            elif k == key.ENTER:
                hiscore.hiscore.add( status.status.score,self.name.element.text,status.status.level_idx )
                director.pop()
                return True
        return False

    def on_text( self, t ):
        if not self.hi_score:
            return False

        if t=='\r':
            return True

        self.name.element.text += t



# ----------------------------------------------------------------------------
# cocos2d
# Copyright (c) 2008 Daniel Moisset, Ricardo Quesada, Rayentray Tappa, Lucio Torre
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
'''Camera Actions

Actions that moves the OpenGL camera.
'''

__docformat__ = 'restructuredtext'

from cocos.director import director
from cocos.euclid import *
from base_actions import *

import math

__all__ = [ 'CameraException',            # Camera Exceptions
           
            'Camera3DAction',
            'OrbitCamera',
            ]

class CameraException( Exception ):
    pass

class Camera3DAction( IntervalAction ):
    def init( self,  duration=2):
        """Initialize the Camera Action

        :Parameters:
            `duration` : int 
                Number of seconds that the action will last
        """
        self.duration = duration

    def start( self ):        
        super(Camera3DAction, self).start()

        self.camera_eye_orig =  self.target.camera.eye
        self.camera_center_orig = self.target.camera.center
        self.camera_up_orig = self.target.camera.up_vector

#    def stop( self ):
#        super(Camera3DAction, self).stop()
#        self.target.camera.eye = self.camera_eye_orig
#        self.target.camera.center = self.camera_center_orig
#        self.target.camera.up_vector = self.camera_up_orig

    def __reversed__( self ):
        return _ReverseTime( self )


class OrbitCamera( Camera3DAction ):
    '''Orbits the camera around the center of the screen using spherical coordinates
    '''
    def init( self, radius=None, delta_radius=0, angle_z=None, delta_z=0, angle_x=None, delta_x=0, *args, **kw ):
        '''Initialize the camera with spherical coordinates

        :Parameters:
            `radius` : float
                Radius of the orbit. Default: current radius
            `delta_radius` : float
                Delta movement of the radius. Default: 0
            `angle_z` : float
                The zenith angle of the spherical coordinate in degrees. Default: current
            `delta_z` : float
                Relative movement of the zenith angle. Default: 0
            `angle_x` : float
                The azimuth angle of the spherical coordinate in degrees. Default: 0
            `delta_x` : float
                Relative movement of the azimuth angle. Default: 0
                

        For more information regarding spherical coordinates, read this:
            http://en.wikipedia.org/wiki/Spherical_coordinates

        '''
        super( OrbitCamera, self ).init( *args, **kw )

        width, height = director.get_window_size()

        self.radius = radius
        self.delta_radius = delta_radius
        self.angle_x = angle_x
        self.rad_delta_x= math.radians( delta_x )
        self.angle_z = angle_z 
        self.rad_delta_z = math.radians( delta_z )

    def start( self ):
        super(OrbitCamera,self).start()

        radius, zenith, azimuth = self.get_spherical_coords()

        if self.radius is None:
            self.radius = radius
        if self.angle_z is None:
            self.angle_z= math.degrees(zenith)
        if self.angle_x is None:
            self.angle_x= math.degrees(azimuth)

        self.rad_x = math.radians( self.angle_x )
        self.rad_z = math.radians( self.angle_z )

    def get_spherical_coords( self ):
        '''returns the spherical coordinates from a cartesian coordinates

        using this formula:

            - http://www.math.montana.edu/frankw/ccp/multiworld/multipleIVP/spherical/body.htm#converting

        :rtype: ( radius, zenith, azimuth )
        '''
        eye = self.target.camera.eye - self.target.camera.center

        radius = math.sqrt( pow(eye.x,2) + pow(eye.y,2) + pow(eye.z,2) )
        s = math.sqrt( pow(eye.x,2) + pow(eye.y,2) )
        if s == 0:
            s = 0.000000001
        r = radius
        if r == 0:
            r = 0.000000001

        angle_z = math.acos( eye.z / float(r) )
        if eye.x < 0:
            angle_x = math.pi - math.asin( eye.y / s )
        else:
            angle_x = math.asin( eye.y / s )

        radius = radius / self.target.camera.get_z_eye()
        return radius, angle_z, angle_x

    def update( self, t ):
        r = (self.radius + self.delta_radius * t) * self.target.camera.get_z_eye()
        z_angle = self.rad_z + self.rad_delta_z * t
        x_angle = self.rad_x + self.rad_delta_x * t

        # using spherical coordinates
        p = Point3( math.sin(z_angle) * math.cos(x_angle),
                    math.sin(z_angle) * math.sin(x_angle),
                    math.cos(z_angle) )

        p = p * r
        d = p + self.camera_center_orig
        self.target.camera.eye = d



# ----------------------------------------------------------------------------
# cocos2d
# Copyright (c) 2008 Daniel Moisset, Ricardo Quesada, Rayentray Tappa, Lucio Torre
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
"""A thin wrapper for OpenGL framebuffer objets. For implementation use only"""

__docformat__ = 'restructuredtext'

from ctypes import c_int, byref
from pyglet.gl import *
from pyglet.gl.glext_arb import *

class FramebufferObject (object):
    """
    Wrapper for framebuffer objects. See

    http://oss.sgi.com/projects/ogl-sample/registry/EXT/framebuffer_object.txt

    API is not very OO, should be improved.
    """
    def __init__ (self):
        """Create a new framebuffer object"""
        id = c_uint(0)
        glGenFramebuffersEXT (1, byref(id))
        self._id = id.value

    def bind (self):
        """Set FBO as current rendering target"""
        glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, self._id)

    def unbind (self):
        """Set default framebuffer as current rendering target"""
        glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0)

    def texture2d (self, texture):
        """Map currently bound framebuffer (not necessarily self) to texture"""
        glFramebufferTexture2DEXT (
            GL_FRAMEBUFFER_EXT,
            GL_COLOR_ATTACHMENT0_EXT,
            texture.target,
            texture.id,
            texture.level)

    def check_status(self):
        """Check that currently set framebuffer is ready for rendering"""
        status = glCheckFramebufferStatusEXT (GL_FRAMEBUFFER_EXT)
        if status != GL_FRAMEBUFFER_COMPLETE_EXT:
            raise Exception ("Frambuffer not complete: %d" % status)

    def __del__(self):
        '''Delete the framebuffer from the GPU memory'''
        id = c_ulong(self._id)
        glDeleteFramebuffersEXT(1, byref(id))



#!/usr/bin/env python

'''Generate PDF documentation

Requires docutils, docbook XSL stylesheets, xsltproc and Apache FOP (with
Batik).
'''

__docformat__ = 'restructuredtext'
__version__ = '$Id: gendoc_pdf.py 1950 2008-03-22 04:07:13Z Alex.Holkner $'

import os
import shutil
import subprocess
from xml.dom.minidom import parse

from docutils.core import publish_doctree, publish_from_doctree
from docutils.writers import docbook
from docutils import nodes

script_home = os.path.dirname(__file__)
doc_root = os.path.join(script_home, '../doc')

def get_docbook_path():
    '''Look up docbook URL in installed catalogs.  If you don't use catalogs,
    hack this function to return the complete path to your docbook dir.
    '''
    url = 'http://docbook.sourceforge.net/release/xsl/current/'
    if 'XML_CATALOG_FILES' in os.environ:
        catalogs = os.environ['XML_CATALOG_FILES'].split()
        for catalog in catalogs:
            proc = subprocess.Popen('xmlcatalog %s %s' % (catalog, url),
                                    shell=True,
                                    stdout=subprocess.PIPE)
            path = proc.communicate()[0]
            if proc.returncode == 0:
                return path.strip()

    raise "Docbook stylesheets not found...  hack me with a path."

def elements(node, name):
    '''Return all descendent element nodes of 'node' with given name'''
    matches = []
    if node.nodeType == node.ELEMENT_NODE and node.nodeName == name:
        matches = [node]
    for child in node.childNodes:
        matches += elements(child, name)
    return matches

def rest2docbook(rest_filename, docbook_filename):
    print 'Reading %s' % rest_filename

    input_dir = os.path.dirname(rest_filename)
    output_dir = os.path.dirname(docbook_filename)

    # Read rest doctree
    doctree = publish_doctree(open(rest_filename).read(),
                              source_path=rest_filename)

    # Remove fields (docbook writer ignores but warns)
    for field in [n for n in doctree.traverse() if isinstance(n, nodes.field)]:
        field.parent.remove(field)

    # Remove line nodes (docbook crashes on them)
    for line in [n for n in doctree.traverse() if isinstance(n, nodes.line)]:
        line.parent.replace(line, line.children)

    # Copy images
    for img in [n for n in doctree.traverse() if isinstance(n, nodes.image)]:
        img['scale'] = '50'
        srcfile = os.path.join(input_dir, img['uri'])
        destfile = os.path.join(output_dir, img['uri'])
        shutil.copyfile(srcfile, destfile)

    print 'Writing %s' % docbook_filename

    # Write docbook xml
    writer = docbook.Writer()
    settings = {
        'doctype': 'book',
    }
    docbook_xml = publish_from_doctree(doctree,     
                                   writer=writer,
                                   settings_overrides=settings)
    open(docbook_filename, 'w').write(docbook_xml)

    # Open docbook xml and fix it
    print 'Reading %s' % docbook_filename
    doc = parse(docbook_filename)

    # Strip leading newline from programlisting
    for elem in elements(doc, 'programlisting'):
        if elem.childNodes and elem.childNodes[0].nodeType == elem.TEXT_NODE:
            elem.childNodes[0].nodeValue = elem.childNodes[0].nodeValue.strip()

    # Dodgy hack to compensate for FOP's lack of table layout.
    # Programming guide tables need more room in the first (header) column than
    # right-hand columns.
    for elem in elements(doc, 'colspec'):
        if elem.getAttribute('colname') == 'col_1':
            elem.attributes['colwidth'] = '2*'
        else:
            elem.attributes['colwidth'] = '1*'

    # Strip table of contents (docbook creates its own)
    for title in elements(doc, 'title'):
        if title.childNodes[0].nodeType == title.TEXT_NODE and \
           title.childNodes[0].nodeValue == 'Contents':
            section = title.parentNode
            if section.nodeType == section.ELEMENT_NODE and \
               section.nodeName == 'section':
                section.parentNode.removeChild(section)

    # Strip local contents
    for section in elements(doc, 'section'):
        for child in section.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                break
        if child.nodeName == 'itemizedlist':
            section.parentNode.removeChild(section)

    # Pull everything before first chapter into a preface
    preface_nodes = []
    preface = doc.createElement('preface')
    preface_title = doc.createElement('title')
    preface_title.appendChild(doc.createTextNode('Welcome'))
    preface.appendChild(preface_title)
    for child in doc.documentElement.childNodes:
        if child.nodeType == child.ELEMENT_NODE:
            if child.nodeName == 'chapter':
                for node in preface_nodes:
                    doc.documentElement.removeChild(node)
                for node in preface_nodes:
                    preface.appendChild(node)
                doc.documentElement.insertBefore(preface, child)
                break
            elif child.nodeName != 'bookinfo':
                preface_nodes.append(child)

    # Scale screenshots of windows down (programming guide hack to fit in
    # table)
    for imagedata in elements(doc, 'imagedata'):
        fileref = imagedata.getAttribute('fileref')
        if fileref.startswith('window_xp_') or fileref.startswith('window_osx'):
            imagedata.attributes['scale'] = '25'

    # Write fixed docbook
    print 'Writing %s' % docbook_filename
    open(docbook_filename, 'w').write(doc.toxml())

def docbook2pdf(docbook_filename, pdf_filename):
    # Run docbook FO stylesheet to produce XSL:FO document
    stylesheet = os.path.join(get_docbook_path(), 'fo/docbook.xsl')
    fo_filename = '%s.fo.xml' % os.path.splitext(docbook_filename)[0] 

    print 'Using %s' % stylesheet
    print 'Writing %s' % fo_filename
    parameters = [
        #'--stringparam paper.type A4',
        #'--param double.sided 1',
        #'--stringparam alignment left',
        #'--param shade.verbatim 1',
        '--param chapter.autolabel 0',
    ]
    result = subprocess.call('xsltproc %s -o %s %s %s ' % (
        ' '.join(parameters), fo_filename, stylesheet, docbook_filename),
        shell=True)

    if result != 0:
        raise 'Aborting due to errors during docbook->FO conversion'

    print 'Writing %s' % pdf_filename
    subprocess.call('fop -fo %s -pdf %s' % (fo_filename, pdf_filename),
                    shell=True)

if __name__ == '__main__':
    try:
        os.makedirs(os.path.join(doc_root, 'pdf'))
    except OSError:
        pass
    try:
        os.makedirs(os.path.join(doc_root, 'docbook'))
    except OSError:
        pass
    rest2docbook(os.path.join(doc_root, 'programming_guide/index.txt'),
                 os.path.join(doc_root, 'docbook/programming_guide.xml'))
    docbook2pdf(os.path.join(doc_root, 'docbook/programming_guide.xml'),
                os.path.join(doc_root, 'pdf/programming_guide.pdf'))



# This code is so you can run the samples without installing the package
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))
#

#!/usr/bin/env python
# -*- coding: utf-8 -*-
import cocos
from cocos.sprite import Sprite
from cocos.actions import *

class TestLayer(cocos.layer.Layer):
    def __init__(self):
        super( TestLayer, self ).__init__()

        sprite1 = Sprite('grossini.png')
        sprite2 = Sprite('grossinis_sister1.png')

        sprite1.add(sprite2)
        self.add(sprite1)
        sprite1.position = (320,240)
        #print sprite1.anchor
        print sprite1.transform_anchor
        print sprite1.children_anchor
        print sprite2.transform_anchor
        print sprite2.children_anchor

if __name__ == '__main__':
    cocos.director.director.init()
    cocos.director.director.run(cocos.scene.Scene(TestLayer()))


#
# original file from http://www.partiallydisassembled.net/make_me/
# modified later for this game
#
from constants import MUSIC, SOUND
import pyglet

try:
    import pyglet.media.avbin
    have_avbin = True
except:
    pyglet.options['audio'] = ('silent')
    have_avbin = False
    MUSIC = False
    SOUND = False

#
# MUSIC
#
music_player = pyglet.media.Player()
current_music = None

sound_vol = 0.7
music_player.volume = 0.4

def set_music(name):
    global current_music

    if not have_avbin:
        return

    if name == current_music:
        return
    current_music = name

    if not MUSIC:
        return

    music_player.next()
    music_player.queue(pyglet.resource.media(name, streaming=True))
    music_player.play()
    # pyglet bug
    music_player.volume = music_player.volume
    music_player.eos_action = 'loop'

def music_volume(vol):
    music_player.volume=vol

def queue_music(name):
    global current_music

    if not have_avbin:
        return

#    if name == current_music:
#        return

    music_player.queue(pyglet.resource.media(name, streaming=True))
    music_player.eos_action = 'next'


def play_music():
    if music_player.playing or not current_music:
        return

    if not have_avbin:
        return

    name = current_music
    music_player.next()
    music_player.queue(pyglet.resource.media(name, streaming=True))
    music_player.play()
    music_player.eos_action = 'loop'

@music_player.event
def on_eos():
    music_player.eos_action = 'loop'


def stop_music():
#    import pdb
#    pdb.set_trace()
    music_player.pause()

#
# SOUND
#
sounds = {}

def load(name, streaming=False):
    if not SOUND:
        return

    if name not in sounds:
        sounds[name] = pyglet.resource.media(name, streaming=streaming)

    return sounds[name]

def play(name):
    if not SOUND:
        return
    load(name)
    a = sounds[name].play().volume = sound_vol

def sound_volume( vol ):
    global sound_vol
    sound_vol = vol



#
# cocos2d
# http://cocos2d.org
#
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from cocos.director import director
from cocos.layer import Layer, ColorLayer
from cocos.scene import Scene
from cocos.scenes.transitions import *
from cocos.actions import *
from cocos.sprite import Sprite
import pyglet
from pyglet import gl, font

from pyglet.window import key

    
class ControlLayer(Layer):

    is_event_handler = True     #: enable pyglet's events

    def __init__( self ):

        super(ControlLayer, self).__init__()

        self.text_title = pyglet.text.Label("Transition Demos",
            font_size=32,
            x=5,
            y=director.get_window_size()[1],
            anchor_x=font.Text.LEFT,
            anchor_y=font.Text.TOP )

        self.text_subtitle = pyglet.text.Label( transition_list[current_transition].__name__,
            font_size=18,
            multiline=True,
            width=600,
            x=5,
            y=director.get_window_size()[1] - 80,
            anchor_x=font.Text.LEFT,
            anchor_y=font.Text.TOP )

        self.text_help = pyglet.text.Label("Press LEFT / RIGHT for prev/next test, ENTER to restart test",
            font_size=16,
            x=director.get_window_size()[0] /2,
            y=20,
            anchor_x=font.Text.CENTER,
            anchor_y=font.Text.CENTER)

    def draw( self ):
        self.text_title.draw()
        self.text_subtitle.draw()
        self.text_help.draw()

    def on_key_press( self, k , m ):
        global current_transition, control_p
        if k == key.LEFT:
            current_transition = (current_transition-1)%len(transition_list)
        elif k == key.RIGHT:
            current_transition = (current_transition+1)%len(transition_list)
        elif k == key.ENTER:
            director.replace( transition_list[current_transition](
                        (control_list[(control_p+1)%len(control_list)] ),
                        1.25)
                    )
            control_p = (control_p + 1) % len(control_list)
            return True

        if k in (key.LEFT, key.RIGHT ):
            self.text_subtitle.text = transition_list[current_transition].__name__


class GrossiniLayer(Layer):
    def __init__( self ):
        super( GrossiniLayer, self ).__init__()

        g = Sprite( 'grossini.png')
        g.position = (320,240)

        rot = RotateBy( 360, 4 )
        g.do( Repeat( rot + Reverse(rot) ) )

        self.add( g )

class GrossiniLayer2(Layer):
    def __init__( self ):
        super( GrossiniLayer2, self ).__init__()

        rot = Rotate( 360, 5 )
        g1 = Sprite( 'grossinis_sister1.png' )
        g1.position = (490,240)

        g2 = Sprite( 'grossinis_sister2.png' )
        g2.position = (140,240)

        g1.do( Repeat( rot + Reverse(rot) ) )
        g2.do( Repeat( rot + Reverse(rot) ) )

        self.add( g1 )
        self.add( g2 )


if __name__ == "__main__":
    director.init(resizable=True)
#    director.window.set_fullscreen(True)

    transition_list = [
        # ActionTransitions
        RotoZoomTransition,
        JumpZoomTransition,
        SplitColsTransition,
        SplitRowsTransition,
        MoveInLTransition,
        MoveInRTransition,
        MoveInBTransition,
        MoveInTTransition,
        SlideInLTransition,
        SlideInRTransition,
        SlideInBTransition,
        SlideInTTransition,
        FlipX3DTransition,
        FlipY3DTransition,
        FlipAngular3DTransition,
        ShuffleTransition,
        ShrinkGrowTransition,
        CornerMoveTransition,
        EnvelopeTransition,
        FadeTRTransition,
        FadeBLTransition,
        FadeUpTransition,
        FadeDownTransition,
        TurnOffTilesTransition,
        FadeTransition,
        ZoomTransition,
        ]
    current_transition = 0

    g = GrossiniLayer()
    g2 = GrossiniLayer2()
    c2 = ColorLayer(128,16,16,255)
    c1 = ColorLayer(0,255,255,255)
    control1 = ControlLayer()
    control2 = ControlLayer()
    controlScene1 = Scene( c2, g, control2 )
    controlScene2 = Scene( c1, g2, control2 )

    control_p = 0
    control_list = [controlScene1, controlScene2]
    
    director.run( controlScene1 )



#!/usr/bin/env python
# $Id:$

import optparse
import os
import os.path
import re
import shutil
import subprocess

from docutils.core import publish_doctree, publish_from_doctree
from docutils.utils import new_document
from docutils.writers import html4css1
from docutils import nodes

from epydoc.markup.doctest import HTMLDoctestColorizer

class HTMLWriter(html4css1.Writer):
    def __init__(self):
        html4css1.Writer.__init__(self)
        self.translator_class = HTMLTranslator

class HTMLTranslator(html4css1.HTMLTranslator):
    def visit_reference(self, node):
        # Copied from html4css1.Writer but extended to add "title" attribute.
        if node.has_key('refuri'):
            href = node['refuri']
            if ( self.settings.cloak_email_addresses
                 and href.startswith('mailto:')):
                href = self.cloak_mailto(href)
                self.in_mailto = 1
        else:
            assert node.has_key('refid'), \
                   'References must have "refuri" or "refid" attribute.'
            href = '#' + node['refid']
        atts = {'href': href, 'class': 'reference'}
        if not isinstance(node.parent, nodes.TextElement):
            assert len(node) == 1 and isinstance(node[0], nodes.image)
            atts['class'] += ' image-reference'
        if node.has_key('link_class'):
            atts['class'] += ' %s' % node['link_class']
        if node.has_key('title'):
            atts['title'] = node['title']
        self.body.append(self.starttag(node, 'a', '', **atts))

def convert_image(uri, input_dir, html_dir):
    '''Convert image found at uri in input_dir to something useable and place
    in html_dir.  Return the new uri.'''

    # Bail if it's external
    if uri.startswith('http:'):
        return uri

    input_file = os.path.join(input_dir, uri)
    if not os.path.exists(input_file):
        print 'Not found: %s' % uri
        return uri

    if uri.endswith('.svg'):
        # Convert SVG to PNG using Inkscape
        newuri = '%s.png' % os.path.splitext(os.path.basename(uri))[0]
        output_file = os.path.join(html_dir, newuri)
        #subprocess.call('convert %s %s' % (input_file, output_file),
        #                shell=True)
        subprocess.call('inkscape --export-png=%s %s' % \
                            (output_file, input_file),
                        shell=True)
        return newuri
    elif uri.endswith('.png') or uri.endswith('.jpg'):
        # Copy image to output dir
        output_file = os.path.join(html_dir, uri)
        shutil.copy(input_file, output_file)
        return uri
        

def shorten(title):
    words = title.split()
    if len(words) <= 3:
        return ' '.join(words)
    else:
        return ' '.join(words[:2]) + ' ...'

class Page(object):
    def __init__(self, document, filename, parent, ids):
        self.document = document
        self.filename = filename
        self.parent = parent
        self.ids = ids
        self.children = []

        # Look for breadcrumb name (title)
        self.title = None
        for field in [n for n in document.traverse() \
                        if isinstance(n, nodes.field)]:
            i = field.first_child_matching_class(nodes.field_name)
            if field[i].astext() == 'breadcrumb':
                j = field.first_child_matching_class(nodes.field_body)
                self.title = field[j].astext()
                field.parent.remove(field)

        # No breadcrumb field, use title
        if self.title is None:
            title_i = document.first_child_matching_class(nodes.Titular)
            if title_i is not None:
                self.title = self.document[title_i].astext()
            else:
                self.title = self.filename
        document['title'] = self.title
 
    def split(self, depth):
        # Remove sections that are children of document and create
        # new pages out of them.
        for section in [n for n in self.document.children \
                        if isinstance(n, nodes.section)]:
            self.document.children.remove(section)

            ids = section.attributes['ids']
            filename = '%s.html' % ids[0].replace('-', '_')
            document = new_document(filename)
            for child in section.children:
                document += child.deepcopy()
            page = Page(document, filename, self, ids)
            self.children.append(page)
            if depth:
                page.split(depth - 1)

    def collect_ids(self, id_map):
        for id in self.ids:
            id_map[id] = self.filename

        for node in self.document.traverse():
            if hasattr(node, 'attributes'):
                for id in node.attributes.get('ids', ()):
                    uri = '%s#%s' % (self.filename, id)
                    id_map[id] = uri

        for child in self.children:
            child.collect_ids(id_map)

    def add_refuri(self, id_map):
        for node in self.document.traverse():
            if hasattr(node, 'attributes'):
                refid = node.attributes.get('refid')
                if refid in id_map:
                    node.attributes['refuri'] = id_map[refid]

        for child in self.children:
            child.add_refuri(id_map)

    def add_navigation(self, previous, next):
        navigation = nodes.container()
        navigation['classes'] += ['navigation']
        paragraph = nodes.paragraph()
        navigation += paragraph

        if previous:
            inline = nodes.inline()
            inline['classes'] += ['previous']
            inline += nodes.Text('Previous: ')
            inline += previous.create_short_reference()
            paragraph += inline

        if next:
            inline = nodes.inline()
            inline['classes'] += ['next']
            inline += nodes.Text('Next: ')
            inline += next.create_short_reference()
            paragraph += inline
        
        breadcrumbs = self.create_short_breadcrumbs()
        if breadcrumbs:
            inline = nodes.inline()
            inline['classes'] += ['breadcrumbs']
            for breadcrumb in breadcrumbs:
                inline += breadcrumb
                inline += nodes.Text(u' \u00bb ')
            inline += nodes.Text(shorten(self.title))
            paragraph += inline

        header = navigation.deepcopy()
        header['classes'] += ['navigation-header']
        self.document.insert(0, header)

        footer = navigation.deepcopy()
        footer['classes'] += ['navigation-footer']
        self.document += footer

    def create_breadcrumbs(self):
        if self.parent:
            return (self.parent.create_breadcrumbs() +
                    [self.parent.create_reference()])
        else:
            return []

    def create_short_breadcrumbs(self):
        if self.parent:
            return (self.parent.create_short_breadcrumbs() +
                    [self.parent.create_short_reference()])
        else:
            return []

    def create_reference(self):
        ref = nodes.reference()
        ref['refuri'] = self.filename
        ref['name'] = self.title
        ref += nodes.Text(self.title)
        return ref

    def create_short_reference(self):
        ref = nodes.reference()
        ref['refuri'] = self.filename
        ref['name'] = self.title
        ref['title'] = self.title
        ref += nodes.Text(shorten(self.title))
        return ref

    def preorder(self):
        yield self
        for child in self.children:
            for c in child.preorder():
                yield c

class PythonColorizer(HTMLDoctestColorizer):
    strong_re = re.compile('(.*)\*\*STRONG_LINE\*\*')

    def colorize_codeblock(self, s):
        text = HTMLDoctestColorizer.colorize_codeblock(self, s)
        text = self.strong_re.sub(self.sub_strong_line, text)
        return text

    def sub_strong_line(self, match):
        return '<strong>%s</strong>' % match.group(1)

    def markup(self, s, tag):
        if tag == 'comment' and s[1:].strip() == '***':
            return '**STRONG_LINE**'
        return HTMLDoctestColorizer.markup(self, s, tag)

def gendoc_html(input_file, html_dir, api_objects, options):
    input_dir = os.path.dirname(input_file)
    files = []
    titles = {}

    # XXX Should calculate how apidoc_dir is relative to html_dir...
    apidoc_dir_rel = '../api' 

    # Read root doctree
    doctree = publish_doctree(open(input_file).read(), source_path=input_file)

    # Convert images
    for image in [n for n in doctree.traverse() if isinstance(n, nodes.image)]:
        uri = image.attributes['uri']
        image.attributes['uri'] = convert_image(uri, input_dir, html_dir)

    # Colorize literal blocks
    for block in [n for n in doctree.traverse() \
                  if (isinstance(n, nodes.literal_block))]:
        pysrc = block.astext()
        html = PythonColorizer().colorize_codeblock(pysrc)
        raw = nodes.raw(text=html, format='html')
        block.replace_self(raw)

    # Recursively split sections down to depth N into separate doctrees
    root_filename = os.path.splitext(os.path.basename(input_file))[0]
    root_page = Page(doctree, 
                     '%s.html' % root_filename,
                     None, [])
    if options.depth:
        root_page.split(options.depth)

    # Add refuri to all references that use refid, to point to the
    # appropriate page.
    id_map = {} # Map id of nodes to uri
    root_page.collect_ids(id_map)
    # Only works for explicit section links; see future uses of id_map
    root_page.add_refuri(id_map) 

    # Get page connectivity and add navigation
    pages = [n for n in root_page.preorder()]
    if options.add_navigation:
        for i, page in enumerate(pages):
            if i > 0:
                previous = pages[i - 1]
            else:
                previous = None
            if i + 1 < len(pages):
                next = pages[i + 1]
            else:
                next = None
            page.add_navigation(previous, next)


    for page in pages:
        # Resolve links
        linked_objects = set()
        for ref in [n for n in page.document.traverse() \
                    if isinstance(n, nodes.title_reference)]:
            title = ref.astext()
            url = None
            if title.endswith('.py'):
                # Copy in referenced example program and link.
                shutil.copy(title, html_dir)
                url = os.path.basename(title)
                canonical = title
                link_class = 'filelink'
            elif title in api_objects and api_objects[title][0]:
                # Link to API page
                canonical, uri = api_objects[title]
                url = os.path.join(apidoc_dir_rel, uri)
                link_class = 'apilink'
            elif title.lower().replace(' ', '-') in id_map:
                # Section link (using `xx` instead of `xx`_).
                canonical = title
                url = id_map[title.lower().replace(' ', '-')]
                link_class = 'sectionlink'

            # Only link once per page, to avoid littering the text
            # with too many links
            if url and url not in linked_objects:
                linked_objects.add(url)

                newref = nodes.reference()
                newref.children = [c.deepcopy() for c in ref.children]
                newref['refuri'] = url
                if canonical != title:
                    newref['title'] = canonical # tooltip is canonical name
                if link_class:
                    newref['link_class'] = link_class
                ref.replace_self(newref)

        # Write page
        settings = {
            'embed_stylesheet': False,
            'stylesheet': 'doc.css',
            'stylesheet_path': None,
        }
        writer = HTMLWriter()
        html = publish_from_doctree(page.document, 
                                    writer=writer,
                                    settings_overrides=settings)
        output_file = open(os.path.join(html_dir, page.filename), 'w')
        output_file.write(html)

def get_api_objects(apidoc_dir):
    '''Return a dict of name -> (canonical, uri) of api_objects exported from
    epydoc.'''
    objects = {}

    # Read API doc objects
    apidoc_file = open(os.path.join(apidoc_dir, 'api-objects.txt'))
    for line in apidoc_file:
        name, url = line.split('\t')

        # Canonical name always matches
        objects[name] = (name, url)
        canonical = name
        # Also strip off leading components one at a time and add if not
        # ambiguous
        while '.' in name:
            name = name.split('.', 1)[1]
            if name in objects and (canonical, url) != objects[name]:
                objects[name] = (None, None) # Ambiguous: don't match
            else:
                objects[name] = (canonical, url)

    return objects


if __name__ == '__main__':
    op = optparse.OptionParser()
    op.add_option('--apidoc-dir')
    op.add_option('--html-dir')
    op.add_option('--depth', type='int')
    op.add_option('--add-navigation', action='store_true')
    options, args = op.parse_args()

    if options.apidoc_dir:
        api_objects = get_api_objects(options.apidoc_dir)
    else:
        api_objects = {}

    for input_file in args:
        gendoc_html(input_file, options.html_dir, api_objects, options)



# ----------------------------------------------------------------------------
# cocos2d
# Copyright (c) 2008 Daniel Moisset, Ricardo Quesada, Rayentray Tappa, Lucio Torre
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
'''Implementation of Grid3DAction actions
'''
__docformat__ = 'restructuredtext'

import math
import random

from cocos.director import director
from cocos.euclid import *
from basegrid_actions import *

rr = random.randrange


__all__ = [
           'Waves3D',   # 3d actions that modifies the z-coordinate
           'FlipX3D',
           'FlipY3D',
           'Lens3D',
           'Shaky3D',
           'Ripple3D',

           'Liquid',    # 3d actions that don't modify the z-coordinate
           'Waves',
           'Twirl',

           ]

class Waves3D( Grid3DAction ):
    '''Simulates waves using the math.sin() function in the z-axis
    The x and y coordinates remains unmodified.

    Example::

       scene.do( Waves3D( waves=5, amplitude=40, grid=(16,16), duration=10) )
    '''

    def init( self, waves=4, amplitude=20, *args, **kw ):
        '''
        :Parameters:
            `waves` : int
                Number of waves (2 * pi) that the action will perform. Default is 4
            `amplitude` : int
                Wave amplitude (height). Default is 20
        '''
        super(Waves3D, self).init( *args, **kw )

        #: Total number of waves to perform
        self.waves=waves
  
        #: amplitude rate. Default: 1.0.
        #: This value is modified by other actions like `AccelAmplitude`.
        self.amplitude_rate = 1.0
        self.amplitude=amplitude

    def update( self, t ):
        for i in xrange(0, self.grid.x+1):
            for j in xrange(0, self.grid.y+1):
                x,y,z = self.get_original_vertex(i,j)

                z += (math.sin(t*math.pi*self.waves*2 + (y+x) * .01) * self.amplitude * self.amplitude_rate )

                self.set_vertex( i,j, (x, y, z) )

 
class FlipX3D( Grid3DAction ):
    '''FlipX3D flips the screen using the Y-axis as a pivot.'''

    def init(self, grid=(1,1), *args, **kw):
        if grid != (1,1):
            raise GridException("Invalid grid size.")
        super(FlipX3D,self).init(grid=grid,*args,**kw)

    def update( self, t ):
        angle = math.pi * t     # 180 degrees
        mz = math.sin( angle )
        angle = angle / 2.0     # x calculates degrees from 0 to 90
        mx = math.cos( angle )

        x0,y,z = self.get_original_vertex(1,1)
        x1,y,z = self.get_original_vertex(0,0)
        
        if x0 > x1:
            # Normal Grid
            a = (0,0)
            b = (0,1)
            c = (1,0)
            d = (1,1)
            x = x0
        else:
            # Reversed Grid
            c = (0,0)
            d = (0,1)
            a = (1,0)
            b = (1,1)
            x = x1

        diff_x = x - x * mx
        diff_z = abs( (x * mz) // 4.0 )
 
        # bottom-left
        x,y,z = self.get_original_vertex(*a)
        self.set_vertex(a[0],a[1],(diff_x,y,z+diff_z))

        # upper-left
        x,y,z = self.get_original_vertex(*b)
        self.set_vertex(b[0],b[1],(diff_x,y,z+diff_z))

        # bottom-right
        x,y,z = self.get_original_vertex(*c)
        self.set_vertex(c[0],c[1],(x-diff_x,y,z-diff_z))

        # upper-right
        x,y,z = self.get_original_vertex(*d)
        self.set_vertex(d[0],d[1],(x-diff_x,y,z-diff_z))
    

class FlipY3D( Grid3DAction ):
    '''FlipY3D flips the screen using the X-axis as a pivot.'''

    def init(self, grid=(1,1), *args, **kw):
        if grid != (1,1):
            raise GridException("Invalid grid size.")
        super(FlipY3D,self).init(grid=grid,*args,**kw)

    def update( self, t ):
        angle = math.pi * t     # 180 degrees
        mz = math.sin( angle )
        angle = angle / 2.0     # x calculates degrees from 0 to 90
        my = math.cos( angle )

        x,y0,z = self.get_original_vertex(1,1)
        x,y1,z = self.get_original_vertex(0,0)
        
        if y0 > y1:
            # Normal Grid
            a = (0,0)
            b = (0,1)
            c = (1,0)
            d = (1,1)
            y = y0
        else:
            # Reversed Grid
            b = (0,0)
            a = (0,1)
            d = (1,0)
            c = (1,1)
            y = y1

        diff_y = y - y * my
        diff_z = abs( (y * mz) // 4.0 )
 
        # bottom-left
        x,y,z = self.get_original_vertex(*a)
        self.set_vertex(a[0],a[1],(x,diff_y,z+diff_z))

        # upper-left
        x,y,z = self.get_original_vertex(*b)
        self.set_vertex(b[0],b[1],(x,y-diff_y,z-diff_z))

        # bottom-right
        x,y,z = self.get_original_vertex(*c)
        self.set_vertex(c[0],c[1],(x,diff_y,z+diff_z))

        # upper-right
        x,y,z = self.get_original_vertex(*d)
        self.set_vertex(d[0],d[1],(x,y-diff_y,z-diff_z))


class Lens3D( Grid3DAction ):
    '''Lens simulates a Lens / Magnifying glass effect.
    It modifies the z-coordinate while the x and y remains unmodified.
    
    Example::

       scene.do( Lens3D(center=(320,240), radius=150, grid=(16,16), duration=10) )
    '''

    def init(self, center=(-1,-1), radius=160, lens_effect=0.7, *args, **kw):
        '''
        :Parameters:
            `center` : (int,int)
                Center of the lens. Default: (win_size_width /2, win_size_height /2 )
            `radius` : int
                Radius of the lens.
            `lens_effect` : float
                How strong is the lens effect. Default: 0.7. 0 is no effect at all, 1 is a very strong lens effect.
        '''
        super(Lens3D,self).init( *args, **kw)
        
        x,y = director.get_window_size()
        if center==(-1,-1):
            center=(x//2, y//2)
            
        #: position of the center of the len. Type: (int,int).
        #: This value can be modified by other actions, like `JumpBy` to simulate a jumping lens
        self.position = Point2( center[0]+1, center[1]+1 )  

        #: radius of the lens. Type: float
        self.radius = radius

        #: lens effect factor. Type: float        
        self.lens_effect = lens_effect
       
        self._last_position = (-1000,-1000)          # dirty vrbl
        
    def update( self, t ):
        if self.position != self._last_position:
            for i in xrange(0, self.grid.x+1):
                for j in xrange(0, self.grid.y+1):
        
                    x,y,z = self.get_original_vertex(i,j)
                    
                    p = Point2( x,y )
                    vect = self.position - p
                    r = abs(vect)
                    
                    if r < self.radius:
        
                        r = self.radius - r
                        pre_log = r/self.radius
                        if pre_log == 0:
                            pre_log = 0.001
                        l = math.log( pre_log )*self.lens_effect
                        new_r = math.exp( l ) * self.radius
        
                        vect.normalize()
                        new_vect = vect * new_r

                        z += abs(new_vect) * self.lens_effect    # magic vrbl        
                    
                    # set all vertex, not only the on the changed
                    # since we want to 'fix' possible moved vertex
                    self.set_vertex( i,j, (x,y,z) )
            self._last_position = self.position


class Ripple3D( Grid3DAction ):
    '''Simulates a ripple (radial wave) effect.
    The amplitude of the wave will decrease when it goes away from the center of the ripple.
    It modifies the z-coordinate while the x and y remains unmodified.
    
    Example::

       scene.do( Ripple3D(center=(320,240), radius=240, waves=15, amplitude=60, duration=20, grid=(32,24) ) )
    '''

    def init(self, center=(-1,-1), radius=240, waves=15, amplitude=60, *args, **kw):
        '''
        :Parameters:
            `center` : (int,int)
                Center of the ripple. Default: (win_size_width /2, win_size_height /2 )
            `radius` : int
                Radius of the ripple. Default: 240
            `waves` : int
                Number of waves (2 * pi) that the action will perform. Default: 15
            `amplitude` : int
                Wave amplitude (height). Default is 60
        '''
        super(Ripple3D,self).init( *args, **kw)
        
        x,y = director.get_window_size()
        if center==(-1,-1):
            center=(x//2, y//2)
            
        #: Center of the ripple. Type: (int,int).
        #: This value can be modified by other actions, like `JumpBy` to simulate a jumping ripple effect
        self.position = Point2( center[0]+1, center[1]+1 )  

        #: radius of the ripple. Type: float
        self.radius = radius

        #: number of waves. Type: int
        self.waves = waves

        #: amplitude rate. Default: 1.0.
        #: This value is modified by other actions like `AccelAmplitude`.
        self.amplitude_rate = 1.0
        self.amplitude=amplitude
       
        
    def update( self, t ):
        for i in xrange(0, self.grid.x+1):
            for j in xrange(0, self.grid.y+1):
    
                x,y,z = self.get_original_vertex(i,j)
                
                p = Point2( x,y )
                vect = self.position - p
                r = abs(vect)
                
                if r < self.radius:
                    r = self.radius - r
                    rate = pow( r / self.radius, 2)
                    z += (math.sin( t*math.pi*self.waves*2 + r * 0.1) * self.amplitude * self.amplitude_rate * rate )
    
                self.set_vertex( i,j, (x,y,z) )


class Shaky3D( Grid3DAction):
    '''Shaky simulates an earthquake by modifying randomly the x, y and z coordinates of each vertex.

    Example::

       scene.do( Shaky3D( randrange=6, grid=(4,4), duration=10) )
    '''
    def init( self, randrange=6, *args, **kw ):
        '''
        :Parameters:
            `randrange` : int
                Number that will be used in random.randrange( -randrange, randrange) to do the effect
        '''
        super(Shaky3D,self).init(*args,**kw)

        #: random range of the shaky effect
        self.randrange = randrange

    def update( self, t ):
        for i in xrange(0, self.grid.x+1):
            for j in xrange(0, self.grid.y+1):
                x,y,z = self.get_original_vertex(i,j)
                x += rr( -self.randrange, self.randrange+1 )
                y += rr( -self.randrange, self.randrange+1 )
                z += rr( -self.randrange, self.randrange+1 )

                self.set_vertex( i,j, (x,y,z) )

class Liquid( Grid3DAction ):
    '''Simulates a liquid effect using the math.sin() function modifying the x and y coordinates.
    The z coordinate remains unmodified.
    
    Example::

       scene.do( Liquid( waves=5, amplitude=40, grid=(16,16), duration=10) )
    '''
    def init( self, waves=4, amplitude=20, *args, **kw ):
        '''
        :Parameters:
            `waves` : int
                Number of waves (2 * pi) that the action will perform. Default is 4
            `amplitude` : int
                Wave amplitude (height). Default is 20
        '''
        super(Liquid, self).init( *args, **kw )

        #: total number of waves
        self.waves=waves

        #: amplitude of the waves
        self.amplitude=amplitude

        #: amplitude rate. Default: 1.0.
        #: This value is modified by other actions like `AccelAmplitude`.
        self.amplitude_rate = 1.0

    def update( self, t ):
            
        for i in xrange(1, self.grid.x):
            for j in xrange(1, self.grid.y):
                x,y,z = self.get_original_vertex(i,j)
                xpos = (x + (math.sin(t*math.pi*self.waves*2 + x * .01) * self.amplitude * self.amplitude_rate))
                ypos = (y + (math.sin(t*math.pi*self.waves*2 + y * .01) * self.amplitude * self.amplitude_rate)) 
                self.set_vertex( i,j, (xpos,ypos,z) )


class Waves( Grid3DAction ):
    '''Simulates waves using the math.sin() function both in the vertical and horizontal axis.
    The z coordinate is not modified.

    Example::

        scene.do( Waves( waves=4, amplitude=20, hsin=False, vsin=True, grid=(16,16), duration=10) )
    '''

    def init( self, waves=4, amplitude=20, hsin=True, vsin=True, *args, **kw ):
        '''Initializes the Waves actions

        :Parameters:
            `waves` : int
                Number of waves (2 * pi) that the action will perform. Default is 4
            `amplitude` : int
                Wave amplitude (height). Default is 20
            `hsin` : bool
                whether or not in will perform horizontal waves. Default is True
            `vsin` : bool
                whether or not in will perform vertical waves. Default is True
        '''
        super(Waves, self).init( *args, **kw )

        #: whether or not it will do horizontal waves
        self.hsin = hsin

        #: whether or not it will do vertical waves
        self.vsin = vsin

        #: total number of wave
        self.waves=waves

        #: amplitude of the waves
        self.amplitude=amplitude

        #: amplitude rate. Default: 1.0
        #: This value is modified by other actions like `AccelAmplitude`.
        self.amplitude_rate = 1.0

    def update( self, t ):        
        for i in xrange(0, self.grid.x+1):
            for j in xrange(0, self.grid.y+1):
                x,y,z = self.get_original_vertex(i,j)
                if self.vsin:
                    xpos = (x + (math.sin(t*math.pi*self.waves*2 + y * .01) * self.amplitude * self.amplitude_rate))
                else:
                    xpos = x

                if self.hsin:
                    ypos = (y + (math.sin(t*math.pi*self.waves*2 + x * .01) * self.amplitude * self.amplitude_rate)) 
                else:
                    ypos = y

                self.set_vertex( i,j, (xpos,ypos,z) )

class Twirl( Grid3DAction ):
    '''Simulates a twirl effect modifying the x and y coordinates.
    The z coordinate is not modified.

    Example::

       scene.do( Twirl( center=(320,240), twirls=5, amplitude=1, grid=(16,12), duration=10) )
    '''
    def init( self, center=(-1,-1), twirls=4, amplitude=1, *args, **kw ):
        '''
        :Parameters:
            `twirls` : int
                Number of twirls (2 * pi) that the action will perform. Default is 4
            `amplitude` : flaot
                Twirl amplitude. Default is 1
            `center` : (int,int)
                Center of the twirl in x,y coordinates. Default: center of the screen
        '''
        super(Twirl, self).init( *args, **kw )

        x,y = director.get_window_size()
        if center==(-1,-1):
            center=(x//2, y//2)
            
        #: position of the center of the Twril. Type: (int,int).
        #: This value can be modified by other actions, like `JumpBy` to simulate a jumping Twirl
        self.position = Point2( center[0]+1, center[1]+1 )  

        #: total number of twirls 
        self.twirls = twirls

        #: amplitude of the twirls
        self.amplitude=amplitude

        #: amplitude rate. Default: 1.0.
        #: This value is modified by other actions like `AccelAmplitude`.
        self.amplitude_rate = 1.0

    def update( self, t ):
            
        cx = self.position.x
        cy = self.position.y

        for i in xrange(0, self.grid.x+1):
            for j in xrange(0, self.grid.y+1):
                x,y,z = self.get_original_vertex(i,j)

                r = math.sqrt( (i-self.grid.x/2.0) ** 2 + (j-self.grid.y/2.0) ** 2 )

                amplitude = 0.1 * self.amplitude * self.amplitude_rate

                a = r * math.cos( math.pi/2.0 + t * math.pi * self.twirls * 2 ) * amplitude

                dx = math.sin(a) * (y-cy) + math.cos(a) * (x-cx)
                dy = math.cos(a) * (y-cy) - math.sin(a) * (x-cx)

                self.set_vertex( i,j, (cx+dx, cy+dy,z) )





import pyglet
from pyglet.gl import *

window = pyglet.window.Window()

grossini = pyglet.resource.image("grossini.png")
grossini.anchor_x = grossini.width / 2
grossini.anchor_y = grossini.height / 2

class MySpriteGroup(pyglet.graphics.Group):
    def __init__(self, sprite):
        super(MySpriteGroup, self).__init__(parent=sprite.group)
        self.sprite = sprite
        
    def set_state(self):
        glPushMatrix()
        self.transform()
        
    def unset_state(self):
        glPopMatrix()
        
    def transform( self ):
        """Apply ModelView transformations"""
        if self.sprite.position != (0,0):
            glTranslatef( self.sprite.position[0], self.sprite.position[1], 0 )

        if self.sprite.scale != 1.0:
            glScalef( self.sprite.scale, self.sprite.scale, 1)

        if self.sprite.rotation != 0.0:
            glRotatef( -self.sprite.rotation, 0, 0, 1)
            
            
batch = pyglet.graphics.Batch()
s1 = pyglet.sprite.Sprite( grossini )
s1.batch = batch
s1.position=(320,100)
g1 = MySpriteGroup(s1)

s2 = pyglet.sprite.Sprite( grossini )
s2.batch = batch
s2.group = g1

s2.position=(00,100)
g2 = MySpriteGroup(s2)
s3 = pyglet.sprite.Sprite( grossini )
s3.batch = batch
s3.group = g2

s3.position=(00,100)


def update(dt):
    s1.rotation += 360*dt
    s2.rotation += 360*dt
    s3.rotation += 360*dt
pyglet.clock.schedule_interval(update, 1/60.)

@window.event
def on_draw():
    window.clear()
    batch.draw()
    
pyglet.app.run()


# This code is so you can run the samples without installing the package
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))
#
from pyglet.gl import *

import cocos
from cocos.director import director
from cocos.sprite import Sprite
import pyglet
from cocos import shader

class TestLayer(cocos.layer.Layer):
    def draw(self):
        x,y = director.get_window_size()
        x = x/2
        y = y/2
        d = 100
        
        cuadric.install()
        glColor4ub(255,255,255,255)
        glBegin(GL_TRIANGLES)
        glTexCoord2f( 0.5,0 )
        glVertex2f( x+d,y+d )
        glTexCoord2f( 0,0 )
        glVertex2f( x,y+d )
        glTexCoord2f( 1,1 )
        glVertex2f( x+d,y )
        
        glTexCoord2f( 0.5,0 )
        glVertex2f( x-d,y-d )
        glTexCoord2f( 0,0 )
        glVertex2f( x,y-d )
        glTexCoord2f( 1,1 )
        glVertex2f( x-d,y )
        
        glTexCoord2f( 0.5,0 )
        glVertex2f( x+d,y-d )
        glTexCoord2f( 0,0 )
        glVertex2f( x,y-d )
        glTexCoord2f( 1,1 )
        glVertex2f( x+d,y )
        
        glTexCoord2f( 0.5,0 )
        glVertex2f( x-d,y+d )
        glTexCoord2f( 0,0 )
        glVertex2f( x,y+d )
        glTexCoord2f( 1,1 )
        glVertex2f( x-d,y )
        
        glTexCoord2f( 0.5,1)
        glVertex2f( x,y )
        glVertex2f( x,y+d )
        glVertex2f( x+d,y )

        glVertex2f( x,y )
        glVertex2f( x,y-d )
        glVertex2f( x+d,y )

        glVertex2f( x,y )
        glVertex2f( x,y+d )
        glVertex2f( x-d,y )

        glVertex2f( x,y )
        glVertex2f( x,y-d )
        glVertex2f( x-d,y )

        glEnd()
        cuadric.uninstall()
                
cuadric_t = '''
void main() {
    vec2 pos = gl_TexCoord[0].st;
    float res = pos.x*pos.x - pos.y;
    if (res<0.0) {
        gl_FragColor = vec4(1.0,1.0,1.0,1.0);
    } else {
        gl_FragColor = vec4(0.0,0.0,0.0,0.0);
    }
}
'''


cuadric = shader.ShaderProgram()
cuadric.setShader(shader.FragmentShader('cuadric_t', cuadric_t))     

if __name__ == "__main__":
    director.init()
    test_layer = TestLayer ()
    main_scene = cocos.scene.Scene (test_layer)
    director.run (main_scene)


#!/usr/bin/python
# $Id:$

'''Rewrite the license header of source files.

Usage:
    license.py file.py file.py dir/ dir/ ...
'''

import optparse
import os
import sys

license = '''# cocos2d
# Copyright (c) 2008 Daniel Moisset, Ricardo Quesada, Rayentray Tappa, Lucio Torre
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.'''

marker = '# ' + '-' * 76

license_lines = [marker] + license.split('\n') + [marker]

def update_license(filename):
    '''Open a Python source file and update the license header, writing
    it back in place.'''
    lines = [l.strip('\r\n') for l in open(filename).readlines()]
    if marker in lines:
        # Update existing license
        try:
            marker1 = lines.index(marker)
            marker2 = lines.index(marker, marker1 + 1)
            if marker in lines[marker2 + 1:]:
                raise ValueError() # too many markers
            lines = (lines[:marker1] + 
                     license_lines + 
                     lines[marker2 + 1:])
        except ValueError:
            print >> sys.stderr, "Can't update license in %s" % filename
    else:
        # Add license to unmarked file
        # Skip over #! if present
        if not lines:
            pass # Skip empty files
        elif lines[0].startswith('#!'):
            lines = lines[:1] + license_lines + lines[1:]
        else:
            lines = license_lines + lines
    open(filename, 'wb').write('\n'.join(lines) + '\n')

if __name__ == '__main__':
    op = optparse.OptionParser()
    op.add_option('--exclude', action='append', default=[])
    options, args = op.parse_args()
    
    if len(args) < 1:
        print >> sys.stderr, __doc__
        sys.exit(0)

    for path in args:
        if os.path.isdir(path):
            for root, dirnames, filenames in os.walk(path):
                for dirname in dirnames:
                    if dirname in options.exclude:
                        dirnames.remove(dirname)
                for filename in filenames:
                    if (filename.endswith('.py') and 
                        filename not in options.exclude):
                        update_license(os.path.join(root, filename))
        else:
            update_license(path)



# This code is so you can run the samples without installing the package
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))
#

import cocos
from cocos.director import director
from cocos.layer import *



if __name__ == "__main__":
    director.init( resizable=True )

    main_scene = cocos.scene.Scene()
    color = ColorLayer(255,0,0,255)
    main_scene.add( color, name='color' )

    # this one doesn't remove it's name
    main_scene.remove(color)

    # this one works
#    main_scene.remove('color')
    main_scene.add( color, name='color' )



# This code is so you can run the samples without installing the package
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../..'))

# stdlib


#pyglet
from pyglet.gl import *

# cocos2d related
from cocos.layer import Layer, ColorLayer
from cocos.scene import Scene
from cocos.director import director
from cocos.actions import *

# tetrico related
from gamectrl import *
from gamemodel import *
import levels
import gameover
from constants import *
import soundex
from HUD import *
from colors import *


__all__ = ['get_newgame']

class GameView( Layer ):

    def __init__(self, model, hud ):
        super(GameView,self).__init__()

        width, height = director.get_window_size()

        aspect = width / float(height)
        self.grid_size = ( int(20 *aspect),20)
        self.duration = 8

        self.position = ( width/2 - COLUMNS * SQUARE_SIZE / 2, 0 )
        self.transform_anchor = ( COLUMNS*SQUARE_SIZE /2, ROWS * SQUARE_SIZE/2)

        # background layer to delimit the pieces visually
        cl = ColorLayer( 112,66,20,50, width = COLUMNS * SQUARE_SIZE, height=ROWS * SQUARE_SIZE )
        self.add( cl, z=-1)

        self.model = model
        self.hud = hud

        self.model.push_handlers( self.on_line_complete, \
                                    self.on_special_effect, \
                                    self.on_game_over, \
                                    self.on_level_complete, \
                                    self.on_move_block, \
                                    self.on_drop_block, \
                                    self.on_new_level, \
                                    self.on_win, \
                                    )

        self.hud.show_message( 'GET READY', self.model.start )

    def on_enter(self):
        super(GameView,self).on_enter()

        soundex.set_music('tetris.mp3')
        soundex.play_music()

    def on_exit(self):
        super(GameView,self).on_exit()
        soundex.stop_music()

    def on_line_complete( self, lines ):
        soundex.play('line.mp3')
        return True

    def on_move_block(self ):
        soundex.play('move.mp3')
        return True

    def on_drop_block(self ):
        soundex.play('drop.mp3')
        return True

    def on_level_complete( self ):
        soundex.play('level_complete.mp3')
        self.hud.show_message('Level complete', self.model.set_next_level)
        return True

    def on_new_level( self ):
        soundex.play('go.mp3')
        self.stop()
        self.do( StopGrid() )
        self.rotation = 0
        self.scale = 1
        return True

    def on_game_over( self ):
        self.parent.add( gameover.GameOver(win=False), z=10 )
        return True

    def on_win( self ):
        self.parent.add( gameover.GameOver(win=True), z=10 )
        return True

    def on_special_effect( self, effects ):
        for e in effects.iterkeys():
            a = self.get_action(e, effects[e])
            self.do( a )

    def get_action( self, e, times=1 ):
        '''returns the actions for a specific effects'''

        w,h= director.get_window_size()
        center = (w/2.0, (ROWS * SQUARE_SIZE) / 2.0  )

        # basic actions
        if e == Colors.ROTATE:
            a = RotateBy( 360 * times, duration=self.duration * times)
        elif e == Colors.SCALE:
            a = ScaleTo( 0.5, duration=self.duration/2.0) + ScaleTo(1, duration=self.duration/2.0) 
            a *= times

        # Grid actions
        elif e == Colors.LIQUID:
            a = Liquid( grid=self.grid_size, waves=self.duration*times // 2, duration=self.duration * times) + StopGrid()
        elif e == Colors.WAVES:
            a = Waves( grid=self.grid_size, waves=self.duration*times // 2, duration=self.duration * times) + StopGrid()
        elif e == Colors.TWIRL:
            a = Twirl( grid=self.grid_size, center=center, twirls=self.duration*times // 2, duration=self.duration * times ) + StopGrid()
        elif e == Colors.LENS:
            w,h = director.get_window_size()
            a = Lens3D( radius=h//2, grid=self.grid_size, duration=self.duration * times) + StopGrid()
        else:
            raise Exception("Effect not implemented: %s" % str(e) )
        return a

    def draw( self ):
        '''draw the map and the block'''

        glPushMatrix()
        self.transform()

        for i in xrange( COLUMNS ):
            for j in xrange( ROWS ):
                color = self.model.map.get( (i,j) )
                if color:
                    Colors.images[color].blit( i * SQUARE_SIZE, j* SQUARE_SIZE)

        if self.model.block:
            self.model.block.draw()

        glPopMatrix()


def get_newgame():
    '''returns the game scene'''
    scene = Scene()

    # model
    model = GameModel()

    # controller
    ctrl = GameCtrl( model )

    # view
    hud = HUD()
    view = GameView( model, hud )

    # set controller in model
    model.set_controller( ctrl )

    # add controller
    scene.add( ctrl, z=1, name="controller" )

    # add view
    scene.add( hud, z=3, name="hud" )
    scene.add( BackgroundLayer(), z=0, name="background" )
    scene.add( view, z=2, name="view" )

    return scene



# ----------------------------------------------------------------------------
# cocos2d
# Copyright (c) 2008 Daniel Moisset, Ricardo Quesada, Rayentray Tappa, Lucio Torre
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
'''Grid data structure'''

__docformat__ = 'restructuredtext'

import pyglet
from pyglet import image
from pyglet.gl import *
from euclid import Point2, Point3

from director import director
import framegrabber

__all__ = ['GridBase',
           'Grid3D',
           'TiledGrid3D',
            ]

class GridBase(object):
    """
    A Scene that takes two scenes and makes a transition between them
    """
    texture = None
    
    def __init__(self):
        super(GridBase, self).__init__()
        self._active = False
        self.reuse_grid = 0     #! Number of times that this grid will be reused

    def init( self, grid ):
        '''Initializes the grid creating both a vertex_list for an independent-tiled grid
        and creating also a vertex_list_indexed for a "united" (non independent tile) grid.

        :Parameters:
            `grid` : euclid.Point2
                size of a 2D grid
        '''
 
        #: size of the grid. (rows, columns)
        self.grid = grid
        
        width, height = director.get_window_size()

        if self.texture is None:
            self.texture = image.Texture.create_for_size(
                    GL_TEXTURE_2D, width, 
                    height, GL_RGBA)
        
        self.grabber = framegrabber.TextureGrabber()
        self.grabber.grab(self.texture)

        #: x pixels between each vertex (float)
        self.x_step = width / self.grid.x
        #: y pixels between each vertex (float)
        self.y_step = height / self.grid.y

        self._init()


    def before_draw( self ):
        '''Binds the framebuffer to a texture
        and set a 2d projection before binding
        to prevent calculating a new texture
        '''

        self._set_2d_projection()

        # capture before drawing
        self.grabber.before_render(self.texture)


    def after_draw( self, camera ):
        '''Called by CocosNode when the texture is already grabbed.
        The FrameBuffer will be unbound and the texture will be drawn

        :Parameters:
            `camera` : `Camera`
                The target's camera object.
        '''

        # capture after drawing
        self.grabber.after_render(self.texture)

        # after unbinding
        # set a 3d projection
        self._set_3d_projection()

        # and center the camera
        camera.locate( force=True )

        # blit
        glEnable(self.texture.target)
        glBindTexture(self.texture.target, self.texture.id)

        glPushAttrib(GL_COLOR_BUFFER_BIT)

        self._blit()
               
        glPopAttrib()
        glDisable(self.texture.target)


    def _set_active(self, bool):
        if self._active == bool:
            return
        self._active = bool
        if self._active == True:
            pass
        elif self._active == False:
            self.vertex_list.delete()
            # to restore the camera to default position
            director.set_projection()
        else:
            raise Exception("Invalid value for GridBase.active")
                                        
    def _get_active(self):
        return self._active

    active = property(_get_active, _set_active,
                      doc='''Determines whether the grid is active or not                 
                     :type: bool
                     ''')       
    def _init(self):
        raise NotImplementedError('abstract')
        
    def _blit(self):
        raise NotImplementedError('abstract')

    def _on_resize(self):
        raise NotImplementedError('abstract')

    @classmethod
    def _set_3d_projection(cls):
        width, height = director.window.width, director.window.height

        glViewport(0, 0, width, height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(60, 1.0*width/height, 0.1, 3000.0)
        glMatrixMode(GL_MODELVIEW)
  
    @classmethod
    def _set_2d_projection(cls):

#        director.set_2d_projection()
        width, height = director.get_window_size()
        glLoadIdentity()
        glViewport(0, 0, width, height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, width, 0, height, -100, 100)
        glMatrixMode(GL_MODELVIEW)
  

class Grid3D(GridBase):
    '''`Grid3D` is a 3D grid implementation. Each vertex has 3 dimensions: x,y,z
    
    The vindexed ertex array will be built with::

        self.vertex_list.vertices: x,y,z (floats)   
        self.vertex_list.tex_coords: x,y,z (floats)
        self.vertex_list.colors: RGBA, with values from 0 - 255
    '''

    def _init( self ):
        # calculate vertex, textures depending on screen size
        idx_pts, ver_pts_idx, tex_pts_idx = self._calculate_vertex_points()

        #: indexed vertex array that can be transformed.
        #: it has these attributes:
        #:
        #:    - vertices
        #:    - colors
        #:    - tex_coords
        #:
        #: for more information refer to pyglet's documentation: pyglet.graphics.vertex_list_indexed
        self.vertex_list = pyglet.graphics.vertex_list_indexed( (self.grid.x+1) * (self.grid.y+1), 
                            idx_pts, "t2f", "v3f/stream","c4B")
        
        #: original vertex array of the grid. (read-only)
        self.vertex_points = ver_pts_idx[:]
        self.vertex_list.vertices = ver_pts_idx
        self.vertex_list.tex_coords = tex_pts_idx
        self.vertex_list.colors = (255,255,255,255) * (self.grid.x+1) * (self.grid.y+1)
 
    def _blit(self ):
        self.vertex_list.draw(pyglet.gl.GL_TRIANGLES)

    def _calculate_vertex_points(self):        
        w = float(self.texture.width)
        h = float(self.texture.height)

        index_points = []
        vertex_points_idx = []
        texture_points_idx = []

        for x in xrange(0,self.grid.x+1):
            for y in xrange(0,self.grid.y+1):
                vertex_points_idx += [-1,-1,-1]
                texture_points_idx += [-1,-1]

        for x in xrange(0, self.grid.x):
            for y in xrange(0, self.grid.y):
                x1 = x * self.x_step 
                x2 = x1 + self.x_step
                y1 = y * self.y_step
                y2 = y1 + self.y_step
              
                #  d <-- c
                #        ^
                #        |
                #  a --> b 
                a = x * (self.grid.y+1) + y
                b = (x+1) * (self.grid.y+1) + y
                c = (x+1) * (self.grid.y+1) + (y+1)
                d = x * (self.grid.y+1) + (y+1)

                # 2 triangles: a-b-d, b-c-d
                index_points += [ a, b, d, b, c, d]    # triangles 

                l1 = ( a*3, b*3, c*3, d*3 )
                l2 = ( Point3(x1,y1,0), Point3(x2,y1,0), Point3(x2,y2,0), Point3(x1,y2,0) )

                #  building the vertex
                for i in xrange( len(l1) ):
                    vertex_points_idx[ l1[i] ] = l2[i].x
                    vertex_points_idx[ l1[i] + 1 ] = l2[i].y
                    vertex_points_idx[ l1[i] + 2 ] = l2[i].z

                # building the texels
                tex1 = ( a*2, b*2, c*2, d*2 )
                tex2 = ( Point2(x1,y1), Point2(x2,y1), Point2(x2,y2), Point2(x1,y2) )

                for i in xrange( len(tex1)):
                    texture_points_idx[ tex1[i] ] = tex2[i].x / w
                    texture_points_idx[ tex1[i] + 1 ] = tex2[i].y / h
 
        return ( index_points, vertex_points_idx, texture_points_idx )

    def get_vertex( self, x, y):
        '''Get the current vertex coordinate

        :Parameters:
            `x` : int 
               x-vertex
            `y` : int
               y-vertex

        :rtype: (float, float, float)
        '''
        idx = (x * (self.grid.y+1) + y) * 3
        x = self.vertex_list.vertices[idx]
        y = self.vertex_list.vertices[idx+1]
        z = self.vertex_list.vertices[idx+2]
        return (x,y,z)

    def get_original_vertex( self, x, y):
        '''Get the original vertex coordinate.
        The original vertices are the ones weren't modified by the current action.

        :Parameters:
            `x` : int 
               x-vertex
            `y` : int
               y-vertex

        :rtype: (float, float, float)
        '''
        idx = (x * (self.grid.y+1) + y) * 3
        x = self.vertex_points[idx]
        y = self.vertex_points[idx+1]
        z = self.vertex_points[idx+2]

        return (x,y,z)


    def set_vertex( self, x, y, v):
        '''Set a vertex point is a certain value

        :Parameters:
            `x` : int 
               x-vertex
            `y` : int
               y-vertex
            `v` : (float, float, float)
                tuple value for the vertex
        '''
        idx = (x * (self.grid.y+1) + y) * 3
        self.vertex_list.vertices[idx] = int(v[0])
        self.vertex_list.vertices[idx+1] = int(v[1])
        self.vertex_list.vertices[idx+2] = int(v[2])

class TiledGrid3D(GridBase):
    '''`TiledGrid3D` is a 3D grid implementation. It differs from `Grid3D` in that
    the tiles can be separated from the grid. 

    The vertex array will be built with::

        self.vertex_list.vertices: x,y,z (floats)   
        self.vertex_list.tex_coords: x,y (floats)
        self.vertex_list.colors: RGBA, with values from 0 - 255
    '''
    def _init( self ):
        # calculate vertex, textures depending on screen size
        ver_pts, tex_pts = self._calculate_vertex_points()

        #: vertex array that can be transformed.
        #: it has these attributes:
        #:
        #:    - vertices
        #:    - colors
        #:    - tex_coords
        #:
        #: for more information refer to pyglet's documentation: pyglet.graphics.vertex_list
        self.vertex_list = pyglet.graphics.vertex_list(self.grid.x * self.grid.y * 4,
                            "t2f", "v3f/stream","c4B")
        #: original vertex array of the grid. (read-only)
        self.vertex_points = ver_pts[:]
        self.vertex_list.vertices = ver_pts
        self.vertex_list.tex_coords = tex_pts
        self.vertex_list.colors = (255,255,255,255) * self.grid.x * self.grid.y * 4  

    def _blit(self ):
        self.vertex_list.draw(pyglet.gl.GL_QUADS)
        
    def _calculate_vertex_points(self):
        w = float(self.texture.width)
        h = float(self.texture.height)

        vertex_points = []
        texture_points = []

        for x in xrange(0, self.grid.x):
            for y in xrange(0, self.grid.y):
                x1 = x * self.x_step 
                x2 = x1 + self.x_step
                y1 = y * self.y_step
                y2 = y1 + self.y_step
              
                # Building the tiles' vertex and texture points
                vertex_points += [x1, y1, 0, x2, y1, 0, x2, y2, 0, x1, y2, 0 ]
                texture_points += [x1/w, y1/h, x2/w, y1/h, x2/w, y2/h, x1/w, y2/h]

        # Generates a quad for each tile, to perform tiles effect
        return (vertex_points, texture_points)

    def set_tile(self, x, y, coords):
        '''Set the 4 tile coordinates

        Coordinates positions::

            3 <-- 2
                  ^
                  |
            0 --> 1

        :Parameters:
            `x` : int 
                x coodinate of the tile
            `y` : int 
                y coordinate of the tile
            `coords` : [ float, float, float, float, float, float, float, float, float, float, float, float ]
                The 4 coordinates in the format (x0, y0, z0, x1, y1, z1,..., x3, y3, z3)
        '''
        idx = (self.grid.y * x + y) * 4 * 3
        self.vertex_list.vertices[idx:idx+12] = coords
    
    def get_original_tile(self, x, y):
        '''Get the 4-original tile coordinates.

        Coordinates positions::

            3 <-- 2
                  ^
                  |
            0 --> 1

        :Parameters:
            `x` : int
                x coordinate of the tile
            `y` : int
                y coordinate of the tile

        :rtype: [ float, float, float, float, float, float, float, float, float, float, float, float ]
        :returns: The 4 coordinates with the following order: x0, y0, z0, x1, y1, z1,...,x3, y3, z3
        '''
        idx = (self.grid.y * x + y) * 4 * 3
        return self.vertex_points[idx:idx+12]

    def get_tile(self, x, y):
        '''Get the current tile coordinates.

        Coordinates positions::

            3 <-- 2
                  ^
                  |
            0 --> 1

        :Parameters:
            `x` : int
                x coordinate of the tile
            `y` : int
                y coordinate of the tile

        :rtype: [ float, float, float, float, float, float, float, float, float, float, float, float ]
        :returns: The 4 coordinates with the following order: x0, y0, z0, x1, y1, z1,...,x3, y3, z3
        '''
        idx = (self.grid.y * x + y) * 4 * 3
        return self.vertex_list.vertices[idx:idx+12]



# ----------------------------------------------------------------------------
# cocos2d
# Copyright (c) 2008 Daniel Moisset, Ricardo Quesada, Rayentray Tappa, Lucio Torre
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
'''Interval Action

Interval Actions
================

An interval action is an action that takes place within a certain period of time.
It has an start time, and a finish time. The finish time is the parameter
``duration`` plus the start time.

These `IntervalAction` have some interesting properties, like:

  - They can run normally (default)
  - They can run reversed with the `Reverse` action.
  - They can run with the time altered with the `Accelerate`, `AccelDeccel` and
    `Speed` actions.

For example, you can simulate a Ping Pong effect running the action normally and
then running it again in Reverse mode.

Example::

    ping_pong_action = action + Reverse( action )


Available IntervalActions
=========================

  * `MoveTo`
  * `MoveBy`
  * `JumpTo`
  * `JumpBy`
  * `Bezier`
  * `Blink`
  * `RotateTo`
  * `RotateBy`
  * `ScaleTo`
  * `ScaleBy`
  * `FadeOut`
  * `FadeIn`
  * `FadeTo`
  * `Delay`
  * `RandomDelay`


Modifier actions
================

  * `Accelerate`
  * `AccelDeccel`
  * `Speed`


Examples::

    move = MoveBy( (200,0), duration=5 )  # Moves 200 pixels to the right in 5 seconds.

    move = MoveTo( (320,240), duration=5) # Moves to the pixel (320,240) in 5 seconds

    jump = JumpBy( (320,0), 100, 5, duration=5) # Jumps to the right 320 pixels
                                                # doing 5 jumps of 100 pixels
                                                # of height in 5 seconds

    accel_move = Accelerate(move)               # accelerates action move
'''

__docformat__ = 'restructuredtext'

import random
import copy
import math

from base_actions import *
from cocos.euclid import *

__all__ = [  'Lerp',                            # interpolation
            'MoveTo','MoveBy',                  # movement actions
            'Jump', 'JumpTo', 'JumpBy',
            'Bezier',                    # complex movement actions
            'Rotate',"RotateTo", "RotateBy",    # object rotation
            'ScaleTo','ScaleBy',                # object scale
            'Delay','RandomDelay',              # Delays
            'FadeOut','FadeIn','FadeTo',        # Fades in/out action
            'Blink',                            # Blink action

            'Accelerate','AccelDeccel','Speed', # Time alter actions
            ]

class Lerp( IntervalAction ):
    """
    Interpolate between values for some specified attribute 
    
    """
    def init(self, attrib, start, end, duration):
        """Init method.

        :Parameters:
            `attrib` : string
                The name of the attrbiute where the value is stored
            `start`  : float
                The start value
            `end`    : float
                The end value
            `duration` : float
                Duration time in seconds
        """
        self.attrib = attrib
        self.duration = duration
        self.start_p = start
        self.end_p = end
        self.delta = end-start

    def update(self, t):
        setattr(self.target, self.attrib,
                self.start_p + self.delta * t
                )

    def __reversed__(self):
        return Lerp(self.attrib, self.end_p, self.start_p, self.duration)
        
class RotateBy( IntervalAction ):
    """Rotates a `CocosNode` object clockwise a number of degrees
    by modiying it's rotation attribute.

    Example::

        # rotates the sprite 180 degrees in 2 seconds
        action = RotateBy( 180, 2 )
        sprite.do( action )
    """
    def init(self, angle, duration ):
        """Init method.

        :Parameters:
            `angle` : float
                Degrees that the sprite will be rotated.
                Positive degrees rotates the sprite clockwise.
            `duration` : float
                Duration time in seconds
        """
        self.angle = angle          #: Quantity of degrees to rotate
        self.duration = duration    #: Duration in seconds

    def start( self ):
        self.start_angle = self.target.rotation
        
    def update(self, t):
        self.target.rotation = (self.start_angle + self.angle * t ) % 360
            
    def __reversed__(self):
        return RotateBy(-self.angle, self.duration)

Rotate = RotateBy


class RotateTo( IntervalAction ):
    """Rotates a `CocosNode` object to a certain angle by modifying it's
    rotation attribute.
    The direction will be decided by the shortest angle.

    Example::

        # rotates the sprite to angle 180 in 2 seconds
        action = RotateTo( 180, 2 )
        sprite.do( action )
    """
    def init(self, angle, duration ):
        """Init method.

        :Parameters:
            `angle` : float
                Destination angle in degrees.
            `duration` : float
                Duration time in seconds
        """
        self.angle = angle%360      #: Destination angle in degrees
        self.duration = duration    #: Duration in seconds

    def start( self ):
        ea = self.angle
        sa = self.start_angle = (self.target.rotation%360)
        self.angle = ((ea%360) - (sa%360))
        if self.angle > 180: 
            self.angle = -360+self.angle
        if self.angle < -180: 
            self.angle = 360+self.angle
        
    def update(self, t):
        self.target.rotation = (self.start_angle + self.angle * t ) % 360

    def __reversed__(self):
        return RotateTo(-self.angle, self.duration)
        
class Speed( IntervalAction ):
    """
    Changes the speed of an action, making it take longer (speed>1)
    or less (speed<1)

    Example::

        # rotates the sprite 180 degrees in 1 secondclockwise
        action = Speed( Rotate( 180, 2 ), 2 )
        sprite.do( action )
    """
    def init(self, other, speed ):
        """Init method.

        :Parameters:
            `other` : IntervalAction
                The action that will be affected
            `speed` : float
                The speed change. 1 is no change.
                2 means twice as fast, takes half the time
                0.5 means half as fast, takes double the time
        """
        self.other = other
        self.speed = speed
        self.duration = other.duration/speed

    def start(self):
        self.other.target = self.target
        self.other.start()

    def update(self, t):
        self.other.update( t )

    def __reversed__(self):
        return Speed( Reverse( self.other ), self.speed )

class Accelerate( IntervalAction ):
    """
    Changes the acceleration of an action

    Example::

        # rotates the sprite 180 degrees in 2 seconds clockwise
        # it starts slow and ends fast
        action = Accelerate( Rotate( 180, 2 ), 4 )
        sprite.do( action )
    """
    def init(self, other, rate = 2):
        """Init method.

        :Parameters:
            `other` : IntervalAction
                The action that will be affected
            `rate` : float
                The acceleration rate. 1 is linear.
                the new t is t**rate
        """
        self.other = other
        self.rate = rate
        self.duration = other.duration

    def start(self):
        self.other.target = self.target
        self.other.start()

    def update(self, t):
        self.other.update( t**self.rate )

    def __reversed__(self):
        return Accelerate(Reverse(self.other), 1.0/self.rate)

class AccelDeccel( IntervalAction ):
    """
    Makes an action change the travel speed but retain near normal
    speed at the beginning and ending.

    Example::

        # rotates the sprite 180 degrees in 2 seconds clockwise
        # it starts slow, gets fast and ends slow
        action = AccelDeccel( RotateBy( 180, 2 ) )
        sprite.do( action )
    """
    def init(self, other):
        """Init method.

        :Parameters:
            `other` : IntervalAction
                The action that will be affected
        """
        self.other = other
        self.duration = other.duration

    def start(self):
        self.other.target = self.target
        self.other.start()

    def update(self, t):
        ft = (t-0.5) * 12
        nt = 1./( 1. + math.exp(-ft) )

        self.other.update( nt )


    def __reversed__(self):
        return AccelDeccel( Reverse(self.other) )



class MoveTo( IntervalAction ):
    """Moves a `CocosNode` object to the position x,y. x and y are absolute coordinates
    by modifying it's position attribute.

    Example::

        # Move the sprite to coords x=50, y=10 in 8 seconds

        action = MoveTo( (50,10), 8 )
        sprite.do( action )
    """
    def init(self, dst_coords, duration=5):
        """Init method.

        :Parameters:
            `dst_coords` : (x,y)
                Coordinates where the sprite will be placed at the end of the action
            `duration` : float
                Duration time in seconds
        """

        self.end_position = Point2( *dst_coords )
        self.duration = duration

    def start( self ):
        self.start_position = self.target.position
        self.delta = self.end_position-self.start_position


    def update(self,t):
        self.target.position = self.start_position + self.delta * t

class MoveBy( MoveTo ):
    """Moves a `CocosNode` object x,y pixels by modifying it's 
    position attribute.
    x and y are relative to the position of the object.
    Duration is is seconds.

    Example::

        # Move the sprite 50 pixels to the left in 8 seconds
        action = MoveBy( (-50,0), 8 )
        sprite.do( action )
    """
    def init(self, delta, duration=5):
        """Init method.

        :Parameters:
            `delta` : (x,y)
                Delta coordinates
            `duration` : float
                Duration time in seconds
        """
        self.delta = Point2( *delta )
        self.duration = duration

    def start( self ):
        self.start_position = self.target.position
        self.end_position = self.start_position + self.delta

    def __reversed__(self):
        return MoveBy(-self.delta, self.duration)

class FadeOut( IntervalAction ):
    """Fades out a `CocosNode` object by modifying it's opacity attribute.

    Example::

        action = FadeOut( 2 )
        sprite.do( action )
    """
    def init( self, duration ):
        """Init method.

        :Parameters:
            `duration` : float
                Seconds that it will take to fade
        """
        self.duration = duration

    def update( self, t ):
        self.target.opacity = 255 * (1-t)

    def __reversed__(self):
        return FadeIn( self.duration )

class FadeTo( IntervalAction ):
    """Fades a `CocosNode` object to a specific alpha value by modifying it's opacity attribute.

    Example::

        action = FadeOut( 2 )
        sprite.do( action )
    """
    def init( self, alpha, duration ):
        """Init method.

        :Parameters:
            `alpha` : float
                0-255 value of opacity
            `duration` : float
                Seconds that it will take to fade
        """
        self.alpha = alpha
        self.duration = duration

    def start(self):
        self.start_alpha = self.target.opacity

    def update( self, t ):
        self.target.opacity = self.start_alpha + (
                    self.alpha - self.start_alpha
                    ) * t


class FadeIn( FadeOut):
    """Fades in a `CocosNode` object by modifying it's opacity attribute.

    Example::

        action = FadeIn( 2 )
        sprite.do( action )
    """
    def update( self, t ):
        self.target.opacity = 255 * t

    def __reversed__(self):
        return FadeOut( self.duration )

class ScaleTo(IntervalAction):
    """Scales a `CocosNode` object to a zoom factor by modifying it's scale attribute.

    Example::

        # scales the sprite to 5x in 2 seconds
        action = ScaleTo( 5, 2 )
        sprite.do( action )
    """
    def init(self, scale, duration=5 ):
        """Init method.

        :Parameters:
            `scale` : float
                scale factor
            `duration` : float
                Duration time in seconds
        """
        self.end_scale = scale
        self.duration = duration

    def start( self ):
        self.start_scale = self.target.scale
        self.delta = self.end_scale-self.start_scale

    def update(self, t):
        self.target.scale = self.start_scale + self.delta * t


class ScaleBy(ScaleTo):
    """Scales a `CocosNode` object a zoom factor by modifying it's scale attribute.

    Example::

        # scales the sprite by 5x in 2 seconds
        action = ScaleBy( 5, 2 )
        sprite.do( action )
    """

    def start( self ):
        self.start_scale = self.target.scale
        self.delta =  self.start_scale*self.end_scale - self.start_scale

    def __reversed__(self):
        return ScaleBy( 1.0/self.end_scale, self.duration )


class Blink( IntervalAction ):
    """Blinks a `CocosNode` object by modifying it's visible attribute
    
    Example::

        # Blinks 10 times in 2 seconds
        action = Blink( 10, 2 )
        sprite.do( action )
    """


    def init(self, times, duration):
        """Init method.

        :Parameters:
            `times` : integer
                Number of times to blink
            `duration` : float
                Duration time in seconds
        """
        self.times = times
        self.duration = duration

    def update(self, t):
        slice = 1 / float( self.times )
        m =  t % slice
        self.target.visible = (m  >  slice / 2.0)

    def __reversed__(self):
        return self


class Bezier( IntervalAction ):
    """Moves a `CocosNode` object through a bezier path by modifying it's position attribute.

    Example::

        action = Bezier( bezier_conf.path1, 5 )   # Moves the sprite using the
        sprite.do( action )                       # bezier path 'bezier_conf.path1'
                                                  # in 5 seconds
    """
    def init(self, bezier, duration=5, forward=True):
        """Init method

        :Parameters:
            `bezier` : bezier_configuration instance
                A bezier configuration
            `duration` : float
                Duration time in seconds
        """
        self.duration = duration
        self.bezier = bezier
        self.forward = forward

    def start( self ):
        self.start_position = self.target.position

    def update(self,t):
        if self.forward:
            p = self.bezier.at( t )
        else:
            p = self.bezier.at( 1-t )
        self.target.position = ( self.start_position +Point2( *p ) )

    def __reversed__(self):
        return Bezier(self.bezier, self.duration, not self.forward)

class Jump(IntervalAction):
    """Moves a `CocosNode` object simulating a jump movement by modifying it's position attribute.

    Example::

        action = Jump(50,200, 5, 6)    # Move the sprite 200 pixels to the right
        sprite.do( action )            # in 6 seconds, doing 5 jumps
                                       # of 50 pixels of height
    """

    def init(self, y=150, x=120, jumps=1, duration=5):
        """Init method

        :Parameters:
            `y` : integer
                Height of jumps
            `x` : integer
                horizontal movement relative to the startin position
            `jumps` : integer
                quantity of jumps
            `duration` : float
                Duration time in seconds
        """

        import warnings
        warnings.warn('Deprecated "Jump" action. Consider using JumpBy instead', DeprecationWarning)
        

        self.y = y
        self.x = x
        self.duration = duration
        self.jumps = jumps

    def start( self ):
        self.start_position = self.target.position

    def update(self, t):
        y = int( self.y * abs( math.sin( t * math.pi * self.jumps ) ) )

        x = self.x * t
        self.target.position = self.start_position + Point2(x,y)

    def __reversed__(self):
        return Jump(self.y, -self.x, self.jumps, self.duration)

class JumpBy(IntervalAction):
    """Moves a `CocosNode` object simulating a jump movement by modifying it's position attribute.

    Example::

        # Move the sprite 200 pixels to the right and up
        action = JumpBy((100,100),200, 5, 6)    
        sprite.do( action )            # in 6 seconds, doing 5 jumps
                                       # of 200 pixels of height
    """

    def init(self, position=(0,0), height=100, jumps=1, duration=5):
        """Init method

        :Parameters:
            `position` : integer x integer tuple
                horizontal and vertical movement relative to the 
                starting position
            `height` : integer
                Height of jumps
            `jumps` : integer
                quantity of jumps
            `duration` : float
                Duration time in seconds
        """
        self.position = position
        self.height = height
        self.duration = duration
        self.jumps = jumps

    def start( self ):
        self.start_position = self.target.position
        self.delta = Vector2(*self.position)
        
    def update(self, t):
        y = int( self.height * abs( math.sin( t * math.pi * self.jumps ) ) )
        y += self.delta[1] * t
        x = self.delta[0] * t
        self.target.position = self.start_position + Point2(x,y)
        
    def __reversed__(self):
        return JumpBy( (-self.position[0],-self.position[1]), self.height, self.jumps, self.duration)

class JumpTo(JumpBy):
    """Moves a `CocosNode` object to a position simulating a jump movement by modifying
    it's position attribute.

    Example::

        action = JumpTo(50,200, 5, 6)  # Move the sprite 200 pixels to the right
        sprite.do( action )            # in 6 seconds, doing 5 jumps
                                       # of 50 pixels of height
    """


    def start( self ):
        self.start_position = self.target.position
        self.delta = Vector2(*self.position)-self.start_position
          

class Delay(IntervalAction):
    """Delays the action a certain amount of seconds

   Example::

        action = Delay(2.5)
        sprite.do( action )
    """
    def init(self, delay):
        """Init method

        :Parameters:
            `delay` : float
                Seconds of delay
        """
        self.duration = delay

    def __reversed__(self):
        return self
        

class RandomDelay(Delay):
    """Delays the actions between *min* and *max* seconds

   Example::

        action = RandomDelay(2.5, 4.5)      # delays the action between 2.5 and 4.5 seconds
        sprite.do( action )
    """
    def init(self, low, hi):
        """Init method

        :Parameters:
            `low` : float
                Minimun seconds of delay
            `hi` : float
                Maximun seconds of delay
        """
        self.low = low
        self.hi = hi

    def __deepcopy__(self, memo):
        new = copy.copy(self)
        new.duration = self.low + (random.random() * (self.hi - self.low))
        return new



# ----------------------------------------------------------------------------
# cocos2d
# Copyright (c) 2008 Daniel Moisset, Ricardo Quesada, Rayentray Tappa, Lucio Torre
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of cocos2d nor the names of its
#     contributors may be used to endorse or promote products
#     derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
"""
CocosNode: the basic element of cocos2d
"""

__docformat__ = 'restructuredtext'

import bisect, copy

import pyglet
from pyglet.gl import *

from director import director
from camera import Camera

import weakref


__all__ = ['CocosNode']

class CocosNode(object):
    """
    Cocosnode is the main element. Anything thats gets drawn or contains things that get drawn is a cocosnode.
    The most popular cocosnodes are scenes, layers and sprites.

    The main features of a cocosnode are:
        - They can contain other cocos nodes (add, get, remove, etc)
        - They can schedule periodic callback (schedule, schedule_interval, etc)
        - They can execute actions (do, pause, stop, etc)

    Some cocosnodes provide extra functionality for them or their children.

    Subclassing a cocosnode usually means (one/all) of:
        - overriding __init__ to initialize resources and schedule calbacks
        - create callbacks to handle the advancement of time
        - overriding draw to render the node
    """
    def __init__(self):
        # composition stuff

        #: list of children. each item is (int, child-reference) where int is the z-order
        self.children = []

        #: dictionary that maps children names with children references
        self.children_names = {}

        self._parent = None

        # drawing stuff

        #: x-position of the object relative to its parent's children_anchor_x value.
        #: Default: 0
        self.x = 0

        #: y-position of the object relative to its parent's children_anchor_y value.
        #: Default: 0
        self.y = 0

        #: a float, alters the scale of this node and its children.
        #: Default: 1.0
        self.scale = 1.0

        #: a float, in degrees, alters the rotation of this node and its children.
        #: Default: 0.0
        self.rotation = 0.0

        #: eye, center and up vector for the `Camera`.
        #: gluLookAt() is used with these values.
        #: Default: FOV 60, center of the screen.
        #: IMPORTANT: The camera can perform exactly the same
        #: transformation as ``scale``, ``rotation`` and the
        #: ``x``, ``y`` attributes (with the exception that the
        #: camera can modify also the z-coordinate)
        #: In fact, they all transform the same matrix, so
        #: use either the camera or the other attributes, but not both
        #: since the camera will be overridden by the transformations done
        #: by the other attributes.
        #: You can change the camera manually or by using the `Camera3DAction`
        #: action.
        self.camera = Camera()


        #: offset from (x,0) from where children will have its (0,0) coordinate.
        #: Default: 0
        self.children_anchor_x = 0

        #: offset from (x,y) from where children will have its (0,0) coordinate.
        #: Default: 0
        self.children_anchor_y = 0

        #: offset from (x,0) from where rotation and scale will be applied.
        #: Default: 0
        self.transform_anchor_x = 0

        #: offset from (0,y) from where rotation and scale will be applied.
        #: Default: 0
        self.transform_anchor_y = 0

        #: whether of not the object is visible.
        #: Default: True
        self.visible = True

        #: the grid object for the grid actions.
        #: This can be a `Grid3D` or a `TiledGrid3D` object depending
        #: on the action.
        self.grid = None

        # actions stuff
        #: list of `Action` objects that are running
        self.actions = []

        #: list of `Action` objects to be removed
        self.to_remove = []

        #: whether or not the next frame will be skipped
        self.skip_frame = False

        # schedule stuff
        self.scheduled = False          # deprecated, soon to be removed
        self.scheduled_calls = []       #: list of scheduled callbacks
        self.scheduled_interval_calls = []  #: list of scheduled interval callbacks
        self.is_running = False         #: whether of not the object is running


    def make_property(attr):
        def set_attr():
            def inner(self, value):
                setattr(self, "children_"+attr,value)
                setattr(self, "transform_"+attr,value)
            return inner
        def get_attr():
            def inner(self):
                if getattr(self,"children_"+attr) != getattr(self, "transform_"+attr):
                    raise Exception("no consistent value for "+attr)
                return getattr(self,"children_"+attr)
            return inner
        return property(
            get_attr(),
            set_attr(),
            doc="""a property to get fast access to [transform_|children_]

            :type: (int,int)
            """+attr )

    #: Anchor point of the object.
    #: Children will be added at this point
    #: and transformations like scaling and rotation will use this point
    #: as the center
    anchor = make_property("anchor")

    #: Anchor x value for transformations and adding children
    anchor_x = make_property("anchor_x")

    #: Anchor y value for transformations and adding children
    anchor_y = make_property("anchor_y")

    def make_property(attr):
        def set_attr():
            def inner(self, value):
                setattr(self, attr+"_x",value[0])
                setattr(self, attr+"_y",value[1])
            return inner
        def get_attr(self):
            return getattr(self,attr+"_x"),  getattr(self,attr+"_y")
        return property(
            get_attr,
            set_attr(),
            doc='''a property to get fast access to "+attr+"_[x|y]

            :type: (int,int)
            ''')

    #: Children anchor point.
    #: Children will be added relative to this point
    children_anchor = make_property("children_anchor")

    #: Transformation anchor point.
    #: Transformations like scaling and rotation
    #: will use this point as it's center
    transform_anchor = make_property("transform_anchor")
    del make_property

    def schedule_interval(self, callback, interval, *args, **kwargs):
        """
        Schedule a function to be called every `interval` seconds.

        Specifying an interval of 0 prevents the function from being
        called again (see `schedule` to call a function as often as possible).

        The callback function prototype is the same as for `schedule`.

        :Parameters:
            `callback` : function
                The function to call when the timer lapses.
            `interval` : float
                The number of seconds to wait between each call.

        This function is a wrapper to pyglet.clock.schedule_interval.
        It has the additional benefit that all calllbacks are paused and
        resumed when the node leaves or enters a scene.

        You should not have to schedule things using pyglet by yourself.
        """
        if self.is_running:
            pyglet.clock.schedule_interval(callback, interval, *args, **kwargs)
        self.scheduled_interval_calls.append(
                (callback, interval, args, kwargs)
                )

    def schedule(self, callback, *args, **kwargs):
        """
        Schedule a function to be called every frame.

        The function should have a prototype that includes ``dt`` as the
        first argument, which gives the elapsed time, in seconds, since the
        last clock tick.  Any additional arguments given to this function
        are passed on to the callback::

            def callback(dt, *args, **kwargs):
                pass

        :Parameters:
            `callback` : function
                The function to call each frame.

        This function is a wrapper to pyglet.clock.schedule.
        It has the additional benefit that all calllbacks are paused and
        resumed when the node leaves or enters a scene.

        You should not have to schedule things using pyglet by yourself.
        """
        if self.is_running:
            pyglet.clock.schedule(callback, *args, **kwargs)
        self.scheduled_calls.append(
                (callback, args, kwargs)
                )

    def unschedule(self, callback):
        """
        Remove a function from the schedule.

        If the function appears in the schedule more than once, all occurances
        are removed.  If the function was not scheduled, no error is raised.

        :Parameters:
            `callback` : function
                The function to remove from the schedule.

        This function is a wrapper to pyglet.clock.unschedule.
        It has the additional benefit that all calllbacks are paused and
        resumed when the node leaves or enters a scene.

        You should not unschedule things using pyglet that where scheduled
        by node.schedule/node.schedule_interface.
        """

        total_len = len(self.scheduled_calls + self.scheduled_interval_calls)
        self.scheduled_calls = [
                c for c in self.scheduled_calls if c[0] != callback
                ]
        self.scheduled_interval_calls = [
                c for c in self.scheduled_interval_calls if c[0] != callback
                ]

        if self.is_running:
            pyglet.clock.unschedule( callback )

    def resume_scheduler(self):
        """
        Time will continue/start passing for this node and callbacks
        will be called.
        """
        for c, i, a, k in self.scheduled_interval_calls:
            pyglet.clock.schedule_interval(c, i, *a, **k)
        for c, a, k in self.scheduled_calls:
            pyglet.clock.schedule(c, *a, **k)

    def pause_scheduler(self):
        """
        Time will stop passing for this node and callbacks will
        not be called
        """
        for f in set(
                [ x[0] for x in self.scheduled_interval_calls ] +
                [ x[0] for x in self.scheduled_calls ]
                ):
            pyglet.clock.unschedule(f)
        for arg in self.scheduled_calls:
            pyglet.clock.unschedule(arg[0])

    def _get_parent(self):
        if self._parent is None: return None
        else: return self._parent()

    def _set_parent(self, parent):
        if parent is None: self._parent = None
        else: self._parent = weakref.ref(parent)

    parent = property(_get_parent, _set_parent, doc='''The parent of this object.

    :type: object
    ''')

    def get_ancestor(self, klass):
        """
        Walks the nodes tree upwards until it finds a node of the class `klass`
        or returns None
        """
        if isinstance(self, klass):
            return self
        parent = self.parent
        if parent:
            return parent.get_ancestor( klass )

    def _get_position(self):
        return (self.x, self.y)
    def _set_position(self, (x,y)):
        self.x, self.y = x,y

    position = property(_get_position, _set_position,
                        doc='''The (x, y) coordinates of the object.

    :type: (int, int)
    ''')

    def add(self, child, z=0, name=None ):
        """Adds a child to the container

        :Parameters:
            `child` : object
                object to be added
            `z` : float
                the z index of self
            `name` : str
                Name of the child
        """
        # child must be a subclass of supported_classes
        #if not isinstance( child, self.supported_classes ):
        #    raise TypeError("%s is not instance of: %s" % (type(child), self.supported_classes) )

        if name:
            if name in self.children_names:
                raise Exception("Name already exists: %s" % name )
            self.children_names[ name ] = child

        child.parent = self

        elem = z, child
        bisect.insort( self.children,  elem )
        if self.is_running:
            child.on_enter()
        return self

    def remove( self, obj ):
        """Removes a child from the container given its name or object

        :Parameters:
            `obj` : string or object
                name of the reference to be removed
                or object to be removed
        """
        if isinstance(obj, str):
            if obj in self.children_names:
                child = self.children_names.pop( obj )
                self._remove( child )
            else:
                raise Exception("Child not found: %s" % obj )
        else:
            self._remove(obj)

    def _remove( self, child ):
        l_old = len(self.children)
        self.children = [ (z,c) for (z,c) in self.children if c != child ]

        if l_old == len(self.children):
            raise Exception("Child not found: %s" % str(child) )

        if self.is_running:
            child.on_exit()

    def get_children(self):
        return [ c for (z, c) in self.children ]

    def __contains__(self, child):
        return  c in self.get_children()

    def get( self, name ):
        """Gets a child from the container given its name

        :Parameters:
            `name` : string
                name of the reference to be get
        """
        if name in self.children_names:
            return self.children_names[ name ]
        else:
            raise Exception("Child not found: %s" % name )

    def on_enter( self ):
        """
        Called every time just before the node enters the stage.
        """
        self.is_running = True

        # start actions
        self.resume()
        # resume scheduler
        self.resume_scheduler()

        # propagate
        for c in self.get_children():
            c.on_enter()


    def on_exit( self ):
        """
        Called every time just before the node leaves the stage
        """
        self.is_running = False

        # pause actions
        self.pause()
        # pause callbacks
        self.pause_scheduler()

        # propagate
        for c in self.get_children():
            c.on_exit()


    def transform( self ):
        """
        Apply ModelView transformations
        you will most likely want to wrap calls to this function with
        glPushMatrix/glPopMatrix
        """
        x,y = director.get_window_size()

        if not(self.grid and self.grid.active):
            # only apply the camera if the grid is not active
            # otherwise, the camera will be applied inside the grid
            self.camera.locate()

        if self.transform_anchor != (0,0):
            glTranslatef(
                self.position[0] + self.transform_anchor_x,
                self.position[1] + self.transform_anchor_y,
                 0 )
        elif self.position != (0,0):
            glTranslatef( self.position[0], self.position[1], 0 )


        if self.scale != 1.0:
            glScalef( self.scale, self.scale, 1)

        if self.rotation != 0.0:
            glRotatef( -self.rotation, 0, 0, 1)

        if self.transform_anchor != (0,0):
            if self.transform_anchor != self.children_anchor:
                glTranslatef(
                    self.children_anchor_x - self.transform_anchor_x,
                    self.children_anchor_y - self.transform_anchor_y,
                     0 )
        #elif self.children_anchor != (0,0):
        #    glTranslatef(
        #        self.children_anchor_x,
        #        self.children_anchor_y,
        #         0 )


    def walk(self, callback, collect=None):
        """
        Executes callback on all the subtree starting at self.
        returns a list of all return values that are not none

        :Parameters:
            `callback` : function
                callable, takes a cocosnode as argument
            `collect` : list
                list of visited nodes

        :rtype: list
        :return: the list of not-none return values

        """
        if collect is None:
            collect = []

        r = callback(self)
        if r is not None:
            collect.append( r )

        for node in self.get_children():
            node.walk(callback, collect)

        return collect

    def visit(self):
        '''
        This function *visits* it's children in a recursive
        way.

        It will first *visit* the children that
        that have a z-order value less than 0.

        Then it will call the `draw` method to
        draw itself.

        And finally it will *visit* the rest of the
        children (the ones with a z-value bigger
        or equal than 0)

        Before *visiting* any children it will call
        the `transform` method to apply any possible
        transformation.
        '''

        if not self.visible:
            return

        position = 0

        if self.grid and self.grid.active:
            self.grid.before_draw()

        # we visit all nodes that should be drawn before ourselves
        if self.children and self.children[0][0] < 0:
            glPushMatrix()
            self.transform()
            for z,c in self.children:
                if z >= 0: break
                position += 1
                c.visit()

            glPopMatrix()

        # we draw ourselves
        self.draw()

        # we visit all the remaining nodes, that are over ourselves
        if position < len(self.children):
            glPushMatrix()
            self.transform()
            for z,c in self.children[position:]:
                c.visit()
            glPopMatrix()

        if self.grid and self.grid.active:
            self.grid.after_draw( self.camera )


    def draw(self, *args, **kwargs):
        """
        This is the function you will have to override if you want your
        subclassed to draw something on screen.

        You *must* respect the position, scale, rotation and anchor attributes.
        If you want OpenGL to do the scaling for you, you can::

            def draw(self):
                glPushMatrix()
                self.transform()
                # ... draw ..
                glPopMatrix()
        """
        pass

    def do( self, action, target=None ):
        '''Executes an *action*.
        When the action finished, it will be removed from the sprite's queue.

        :Parameters:
            `action` : an `Action` instance
                Action that will be executed.
        :rtype: `Action` instance
        :return: A clone of *action*
        '''
        a = copy.deepcopy( action )

        if target is None:
            a.target = self
        else:
            a.target = target

        a.start()
        self.actions.append( a )

        if not self.scheduled:
            if self.is_running:
                self.scheduled = True
                pyglet.clock.schedule( self._step )
        return a

    def remove_action(self, action ):
        """Removes an action from the queue

        :Parameters:
            `action` : Action
                Action to be removed
        """
        self.to_remove.append( action )

    def pause(self):
        """
        Suspends the execution of actions.
        """
        if not self.scheduled:
            return
        self.scheduled = False
        pyglet.clock.unschedule( self._step )

    def resume(self):
        """
        Resumes the execution of actions.
        """
        if self.scheduled:
            return
        self.scheduled = True
        pyglet.clock.schedule( self._step )
        self.skip_frame = True

    def stop(self):
        """
        Removes all actions from the running action list
        """
        for action in self.actions:
            self.to_remove.append( action )

    def are_actions_running(self):
        """
        Determine whether any actions are running.
        """
        return bool(set(self.actions) - set(self.to_remove))

    def _step(self, dt):
        """This method is called every frame.

        :Parameters:
            `dt` : delta_time
                The time that elapsed since that last time this functions was called.
        """
        for x in self.to_remove:
            if x in self.actions:
                self.actions.remove( x )
        self.to_remove = []

        if self.skip_frame:
            self.skip_frame = False
            return

        if len( self.actions ) == 0:
            self.scheduled = False
            pyglet.clock.unschedule( self._step )

        for action in self.actions:
            action.step(dt)
            if action.done():
                action.stop()
                self.remove_action( action )





# This code is so you can run the samples without installing the package
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))
#


import cocos
from cocos.director import director
from cocos.sprite import Sprite
import pyglet
        
class PrintKey(cocos.layer.Layer):
    is_event_handler = True
    def on_key_press (self, key, modifiers):
        print "Key Pressed:", key, modifiers

class SwitchLayer(cocos.layer.Layer):
    def __init__(self, subscene):
        super(SwitchLayer, self).__init__()
        self.subscene = subscene
        self.dispatch = False
    
    is_event_handler = True
    def on_key_press (self, key, modifiers):
        if key == pyglet.window.key.SPACE:
            self.dispatch = not self.dispatch
            print "DISPATCH EVENTS:", self.dispatch
            self.subscene.enable_handlers(self.dispatch)
            
    
if __name__ == "__main__":
    director.init()
    bg_layer = cocos.layer.ColorLayer(255,0,0,255)
    sub = cocos.scene.Scene(bg_layer, PrintKey())
    test_layer = SwitchLayer(sub)
    main_scene = cocos.scene.Scene (sub, test_layer)
    director.run (main_scene)



# This code is so you can run the samples without installing the package
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../..'))

# stdlib
import copy
import random
import weakref

# pyglet related
import pyglet

# cocos2d related
from cocos.euclid import Point2

# tetrico related
from constants import *
from status import status
from colors import *
import levels

__all__ = ['GameModel']

#
# Model (of the MVC pattern)
#

class GameModel( pyglet.event.EventDispatcher ):

    def __init__(self):
        super(GameModel,self).__init__()

        self.init()

        # current block
        self.block = None

        # grid
        self.map = {}

        status.reset()


        # phony level
        status.level = levels.levels[0]

    def start( self ):
        self.set_next_level()

    def set_controller( self, ctrl ):
        self.ctrl = weakref.ref( ctrl )

    def init_map(self):
        '''creates a map'''
        self.map= {}
        for i in xrange( COLUMNS ):
            for j in xrange( ROWS ):
                self.map[ (i,j) ] = 0

    def check_line(self):
        '''checks if the line is complete'''
        lines = []
        for j in xrange( ROWS ):
            for i in xrange( COLUMNS ):
                c = self.map.get( (i,j) )
                if not c:
                    break
                if i == COLUMNS-1:
                    lines.append(j)

        lines.reverse()

        effects = []
        for j in lines:
            for i in xrange(COLUMNS):
                e = self.map[ (i,j) ]
                if e in Colors.specials:
                    effects.append( e )

        if effects:
            self.process_effects( effects )

        for l in lines:
            for j in xrange(l, ROWS-1 ):
                for i in xrange(COLUMNS):
                    self.map[ (i,j) ] = self.map[ (i,j+1) ]

        if lines:
            status.score += pow(2, len(lines)) -1
            status.lines += len(lines)
            self.dispatch_event("on_line_complete", lines )

            if status.lines >= status.level.lines:
                self.ctrl().pause_controller()
                if status.level_idx + 1 >= len( levels.levels):
                    self.dispatch_event("on_win")
                else:
                    self.dispatch_event("on_level_complete")

    def init( self ):
        status.lines = 0
        self.init_map()

    def set_next_level( self ):
        self.ctrl().resume_controller()

        if status.level_idx is None:
            status.level_idx = 0
        else:
            status.level_idx += 1


        l = levels.levels[ status.level_idx ]

        self.init()
        status.level = l()

        # eliminate current and next
        self.random_block()
        self.random_block()

        self.dispatch_event("on_new_level")


    def process_effects( self, effects ):
        d = {}
        elements = set(effects)
        for e in elements:
            d[ e ] = effects.count(e)

        self.dispatch_event("on_special_effect", d )

    def merge_block( self ):
        '''merges a block in the map'''
        for i in xrange( self.block.x ):
            for j in xrange( self.block.x ):
                c= self.block.get(i,j)
                if c:
                    self.map[ (i+self.block.pos.x, j+self.block.pos.y) ] = c

    def are_valid_movements(self):
        '''check wheter there are any left valid movement'''
        for i in xrange(self.block.x):
            for j in xrange(self.block.x):
                if self.block.get(i,j):
                    if j + self.block.pos.y == 0:
                        return False
                    if self.map.get( (i + self.block.pos.x,j + self.block.pos.y -1), False ):
                        return False
        return True

    def block_right( self ):
        '''moves right the block 1 square'''
        self.block.backup()
        self.block.pos.x += 1
        if not self.is_valid_block():
            self.block.restore()
        else:
            self.dispatch_event("on_move_block")

    def block_left( self ):
        '''moves left the block 1 square'''
        self.block.backup()
        self.block.pos.x -= 1
        if not self.is_valid_block():
            self.block.restore()
        else:
            self.dispatch_event("on_move_block")

    def block_down( self, sound=True ):
        '''moves down the block 1 square'''
        self.block.backup()
        self.block.pos.y -= 1
        if not self.is_valid_block():
            self.block.restore()
            self.next_block()
        else:
            if sound:
                self.dispatch_event("on_move_block")

    def block_drop( self ):
        '''drops the block'''
        while True:
            self.block.backup()
            self.block.pos.y -= 1
            if not self.is_valid_block():
                self.block.restore()
                break
        self.dispatch_event("on_drop_block")

    def block_rotate( self ):
        '''rotates the block'''
        self.block.backup()
        self.block.rotate()
        if not self.is_valid_block():
            self.block.restore()
        else:
            self.dispatch_event("on_move_block")

    def next_block(self):
        '''merge current block in grid,
        check if there are lines completed,
        and choose a new random block'''
        self.merge_block()
        self.check_line()
        self.random_block()

        if not self.is_valid_block():
            self.ctrl().pause_controller()
            self.dispatch_event("on_game_over")

    def random_block( self ):
        '''puts the next block in stage'''
        self.block = status.next_piece
        block = random.choice( (
            Block_L,
            Block_L2,
            Block_O,
            Block_I,
            Block_Z,
            Block_Z2,
            Block_A
            ) )
        status.next_piece = block()

        if not self.block:
            self.random_block()

    def is_valid_block( self ):
        '''check wheter the block is valid in the current position'''
        for i in xrange( self.block.x ):
            for j in xrange( self.block.x ):
                if self.block.get(i,j):
                    if self.block.pos.x+i < 0:
                        return False
                    if self.block.pos.x+i >= COLUMNS:
                        return False
                    if self.block.pos.y+j < 0:
                        return False
                    if self.map.get( (self.block.pos.x+i, self.block.pos.y+j), False ): 
                        return False
        return True


class Block( object ):
    '''Base class for all blocks'''
    def __init__(self):
        super( Block, self).__init__()

        self.pos = Point2( COLUMNS/2-1, ROWS )
        self.rot = 0

        for x in xrange( len(self._shape) ):
            for y in xrange( len( self._shape[x]) ):
                if self._shape[x][y]:
                    r = random.random()
                    if r < status.level.prob:
                        color = random.choice( status.level.blocks )
                    else:
                        color = self.color
                    self._shape[x][y] = color

    def draw( self ):
        '''draw the block'''
        for i in xrange(self.x):
            for j in xrange(self.x):
                c = self.get(i,j)
                if c:
                    Colors.images[c].blit( (i + self.pos.x) * SQUARE_SIZE, (j + self.pos.y) * SQUARE_SIZE)

    def rotate( self ):
        '''rotate the block'''
        self.rot = (self.rot + 1) % self.mod

    def backup( self ):
        '''saves a copy of the block'''
        self.save_pos = copy.copy( self.pos )
        self.save_rot = self.rot

    def restore( self ):
        '''restore a copy of the block'''
        self.pos = self.save_pos
        self.rot = self.save_rot

    def get(self,x,y):
        '''get position x,y of the block'''
        if self.rot == 0:
            i,j = x,y
        elif self.rot == 1:
            i,j = y,(self.x -x -1 )
        elif self.rot == 2:
            i,j = (self.x - x -1), (self.x -y -1)
        elif self.rot == 3:
            i,j = (self.x - y -1 ), x

        return self._shape[i][j]


class Block_I( Block ):
    x = 4
    mod = 2
    color = Colors.RED

    def __init__(self):
        self._shape = [ [0,1,0,0],
                   [0,1,0,0],
                   [0,1,0,0],
                   [0,1,0,0] ]
        super(Block_I,self).__init__()

class Block_Z( Block ):
    x = 3
    color = Colors.ORANGE
    mod = 2

    def __init__(self):
        self._shape = [ [0,0,0],
                       [1,1,0],
                       [0,1,1] ]
        super(Block_Z,self).__init__()

class Block_Z2( Block ):
    x = 3
    color = Colors.CYAN
    mod = 2

    def __init__(self):
        self._shape = [ [0,0,0],
                       [0,1,1],
                       [1,1,0] ]
        super(Block_Z2,self).__init__()

class Block_O( Block ):
    x = 2
    color = Colors.BLUE
    mod = 4

    def __init__(self):
        self._shape = [ [1,1],
                       [1,1] ]
        super(Block_O,self).__init__()

class Block_L( Block ):
    x = 3
    color = Colors.MAGENTA
    mod = 4

    def __init__(self):
        self._shape = [ [1,0,0],
                       [1,0,0],
                       [1,1,0] ] 
        super(Block_L,self).__init__()

class Block_L2( Block ):
    x = 3
    color = Colors.YELLOW
    mod = 4

    def __init__(self):
        self._shape = [ [0,0,1],
                       [0,0,1],
                       [0,1,1] ] 
        super(Block_L2,self).__init__()

class Block_A( Block ):
    x = 3
    color = Colors.GREEN
    mod = 4

    def __init__(self):
        self._shape = [ [0,0,0],
                       [0,1,0],
                       [1,1,1] ] 
        super(Block_A,self).__init__()

GameModel.register_event_type('on_special_effect')
GameModel.register_event_type('on_line_complete')
GameModel.register_event_type('on_level_complete')
GameModel.register_event_type('on_new_level')
GameModel.register_event_type('on_game_over')
GameModel.register_event_type('on_move_block')
GameModel.register_event_type('on_drop_block')
GameModel.register_event_type('on_win')
